# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/7/29

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The title case capitalization for the base account titles.
This follows the APA style title case capitalization.  See
https://apastyle.apa.org/style-grammar-guidelines/capitalization/title-case .

This module should not import any other module from the application.

"""
import re

CONJUNCTIONS: set[str] = {"and", "as", "but", "for", "if", "nor", "or", "so",
                          "yet"}
"""Short conjunctions."""
ARTICLES: set[str] = {"a", "an", "the"}
"""Articles."""
PREPOSITIONS: set[str] = {"as", "at", "by", "for", "in", "of", "on", "per",
                          "to", "up", "via"}
"""Short prepositions."""
MINOR_WORDS: set[str] \
    = CONJUNCTIONS.copy().union(ARTICLES).union(PREPOSITIONS)
"""Minor words that should be in lowercase."""
# Excludes "by" as in "1223 by-products"
MINOR_WORDS.remove("by")


def title_case(s: str) -> str:
    """Capitalize a title string for the base account titles.  Do not use it
    in other places.  This excludes "by" as in "1223 by-products".

    :param s: The title string.
    :return: The capitalized title string.
    """
    return re.sub(r"\w+", __cap_word, s)


def __cap_word(m: re.Match) -> str:
    """Capitalize a matched title word.

    :param m: The matched title word.
    :return: The capitalized title word.
    """
    if m.group(0).lower() in MINOR_WORDS:
        return m.group(0)
    return m.group(0).title()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the account management.

"""
import datetime as dt
import unittest

import httpx
from flask import Flask

from accounting.utils.next_uri import encode_next
from test_site import db
from testlib import NEXT_URI, create_test_app, get_client, get_csrf_token, \
    set_locale, add_journal_entry


class AccountData:
    """The account data."""

    def __init__(self, base_code: str, no: int, title: str):
        """Constructs the account data.

        :param base_code: The base code.
        :param no: The number.
        :param title: The title.
        """
        self.base_code: str = base_code
        """The base code."""
        self.no: int = no
        """The number."""
        self.title: str = title
        """The title."""
        self.code: str = f"{self.base_code}-{self.no:03d}"
        """The code."""


CASH: AccountData = AccountData("1111", 1, "Cash")
"""The cash account."""
PETTY: AccountData = AccountData("1112", 1, "Bank")
"""The petty cash account."""
BANK: AccountData = AccountData("1113", 1, "Bank")
"""The bank account."""
STOCK: AccountData = AccountData("1121", 1, "Stock")
"""The stock account."""
LOAN: AccountData = AccountData("2112", 1, "Loan")
"""The loan account."""
PREFIX: str = "/accounting/accounts"
"""The URL prefix for the account management."""


class AccountTestCase(unittest.TestCase):
    """The account test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""

        with self.__app.app_context():
            from accounting.models import Account, AccountL10n
            AccountL10n.query.delete()
            Account.query.delete()
            db.session.commit()
            self.__encoded_next_uri: str = encode_next(NEXT_URI)
            """The encoded next URI."""

        self.__client: httpx.Client = get_client(self.__app, "editor")
        """The user client."""
        self.__csrf_token: str = get_csrf_token(self.__client)
        """The CSRF token."""
        response: httpx.Response

        response = self.__client.post(f"{PREFIX}/store",
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": CASH.base_code,
                                            "title": CASH.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{CASH.code}")

        response = self.__client.post(f"{PREFIX}/store",
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": BANK.base_code,
                                            "title": BANK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{BANK.code}")

    def test_nobody(self) -> None:
        """Test the permission as nobody.

        :return: None.
        """
        from accounting.models import Account
        client: httpx.Client = get_client(self.__app, "nobody")
        csrf_token: str = get_csrf_token(client)
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{CASH.code}")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/create")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/store",
                               data={"csrf_token": csrf_token,
                                     "base_code": STOCK.base_code,
                                     "title": STOCK.title})
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{CASH.code}/edit")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{CASH.code}/update",
                               data={"csrf_token": csrf_token,
                                     "base_code": CASH.base_code,
                                     "title": f"{CASH.title}-2"})
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{BANK.code}/delete",
                               data={"csrf_token": csrf_token})
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/bases/{CASH.base_code}")
        self.assertEqual(response.status_code, 403)

        with self.__app.app_context():
            cash_id: int = Account.find_by_code(CASH.code).id

        response = client.post(f"{PREFIX}/bases/{CASH.base_code}",
                               data={"csrf_token": csrf_token,
                                     "next": self.__encoded_next_uri,
                                     f"{cash_id}-no": "5"})
        self.assertEqual(response.status_code, 403)

    def test_viewer(self) -> None:
        """Test the permission as viewer.

        :return: None.
        """
        from accounting.models import Account
        client: httpx.Client = get_client(self.__app, "viewer")
        csrf_token: str = get_csrf_token(client)
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/{CASH.code}")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/create")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/store",
                               data={"csrf_token": csrf_token,
                                     "base_code": STOCK.base_code,
                                     "title": STOCK.title})
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{CASH.code}/edit")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{CASH.code}/update",
                               data={"csrf_token": csrf_token,
                                     "base_code": CASH.base_code,
                                     "title": f"{CASH.title}-2"})
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{BANK.code}/delete",
                               data={"csrf_token": csrf_token})
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/bases/{CASH.base_code}")
        self.assertEqual(response.status_code, 200)

        with self.__app.app_context():
            cash_id: int = Account.find_by_code(CASH.code).id

        response = client.post(f"{PREFIX}/bases/{CASH.base_code}",
                               data={"csrf_token": csrf_token,
                                     "next": self.__encoded_next_uri,
                                     f"{cash_id}-no": "5"})
        self.assertEqual(response.status_code, 403)

    def test_editor(self) -> None:
        """Test the permission as editor.

        :return: None.
        """
        from accounting.models import Account
        response: httpx.Response

        response = self.__client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/{CASH.code}")
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(f"{PREFIX}/create")
        self.assertEqual(response.status_code, 200)

        response = self.__client.post(f"{PREFIX}/store",
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": STOCK.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{STOCK.code}")

        response = self.__client.get(f"{PREFIX}/{CASH.code}/edit")
        self.assertEqual(response.status_code, 200)

        response = self.__client.post(f"{PREFIX}/{CASH.code}/update",
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": CASH.base_code,
                                            "title": f"{CASH.title}-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{CASH.code}")

        response = self.__client.post(f"{PREFIX}/{BANK.code}/delete",
                                      data={"csrf_token": self.__csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], PREFIX)

        response = self.__client.get(f"{PREFIX}/bases/{CASH.base_code}")
        self.assertEqual(response.status_code, 200)

        with self.__app.app_context():
            cash_id: int = Account.find_by_code(CASH.code).id

        response = self.__client.post(f"{PREFIX}/bases/{CASH.base_code}",
                                      data={"csrf_token": self.__csrf_token,
                                            "next": self.__encoded_next_uri,
                                            f"{cash_id}-no": "5"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], NEXT_URI)

    def test_add(self) -> None:
        """Tests to add the currencies.

        :return: None.
        """
        from accounting.models import Account
        create_uri: str = f"{PREFIX}/create"
        store_uri: str = f"{PREFIX}/store"
        detail_uri: str = f"{PREFIX}/{STOCK.code}"
        response: httpx.Response

        with self.__app.app_context():
            self.assertEqual({x.code for x in Account.query.all()},
                             {CASH.code, BANK.code})

        # Missing CSRF token
        response = self.__client.post(store_uri,
                                      data={"base_code": STOCK.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 400)

        # CSRF token mismatch
        response = self.__client.post(store_uri,
                                      data={"csrf_token":
                                            f"{self.__csrf_token}-2",
                                            "base_code": STOCK.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 400)

        # Empty base account code
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": " ",
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Non-existing base account
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "9999",
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Unavailable base account
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "1",
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Empty name
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": STOCK.base_code,
                                            "title": " "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # A nominal account that needs offset
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "6172",
                                            "title": STOCK.title,
                                            "is_need_offset": "yes"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Success, with spaces to be stripped
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": f" {STOCK.base_code} ",
                                            "title": f" {STOCK.title} "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        # Success under the same base
        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": STOCK.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{STOCK.base_code}-002")

        # Success under the same base, with order in a mess.
        with self.__app.app_context():
            stock_2: Account = Account.find_by_code(f"{STOCK.base_code}-002")
            stock_2.no = 66
            db.session.commit()

        response = self.__client.post(store_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": STOCK.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{STOCK.base_code}-003")

        with self.__app.app_context():
            self.assertEqual({x.code for x in Account.query.all()},
                             {CASH.code, BANK.code, STOCK.code,
                              f"{STOCK.base_code}-002",
                              f"{STOCK.base_code}-003"})

            account: Account = Account.find_by_code(STOCK.code)
            self.assertEqual(account.base_code, STOCK.base_code)
            self.assertEqual(account.title_l10n, STOCK.title)

    def test_basic_update(self) -> None:
        """Tests the basic rules to update a user.

        :return: None.
        """
        from accounting.models import Account
        detail_uri: str = f"{PREFIX}/{CASH.code}"
        edit_uri: str = f"{PREFIX}/{CASH.code}/edit"
        update_uri: str = f"{PREFIX}/{CASH.code}/update"
        detail_c_uri: str = f"{PREFIX}/{STOCK.code}"
        response: httpx.Response

        # Success, with spaces to be stripped
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": f" {CASH.base_code} ",
                                            "title": f" {CASH.title}-1 "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account: Account = Account.find_by_code(CASH.code)
            self.assertEqual(account.base_code, CASH.base_code)
            self.assertEqual(account.title_l10n, f"{CASH.title}-1")

        # Empty base account code
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": " ",
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Non-existing base account
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "9999",
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Unavailable base account
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "1",
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Empty name
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": STOCK.base_code,
                                            "title": " "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # A nominal account that needs offset
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "6172",
                                            "title": STOCK.title,
                                            "is_need_offset": "yes"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Change the base account
        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": STOCK.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_c_uri)

        response = self.__client.get(detail_uri)
        self.assertEqual(response.status_code, 404)

        response = self.__client.get(detail_c_uri)
        self.assertEqual(response.status_code, 200)

    def test_update_not_modified(self) -> None:
        """Tests that the data is not modified.

        :return: None.
        """
        from accounting.models import Account
        detail_uri: str = f"{PREFIX}/{CASH.code}"
        update_uri: str = f"{PREFIX}/{CASH.code}/update"
        account: Account
        response: httpx.Response

        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": f" {CASH.base_code} ",
                                            "title": f" {CASH.title} "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertIsNotNone(account)
            account.created_at \
                = account.created_at - dt.timedelta(seconds=5)
            account.updated_at = account.created_at
            db.session.commit()

        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": CASH.base_code,
                                            "title": STOCK.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertIsNotNone(account)
            self.assertLess(account.created_at,
                            account.updated_at)

    def test_created_updated_by(self) -> None:
        """Tests the created-by and updated-by record.

        :return: None.
        """
        from accounting.models import Account
        editor_username, admin_username = "editor", "admin"
        client: httpx.Client = get_client(self.__app, admin_username)
        csrf_token: str = get_csrf_token(client)
        detail_uri: str = f"{PREFIX}/{CASH.code}"
        update_uri: str = f"{PREFIX}/{CASH.code}/update"
        account: Account
        response: httpx.Response

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertEqual(account.created_by.username, editor_username)
            self.assertEqual(account.updated_by.username, editor_username)

        response = client.post(update_uri,
                               data={"csrf_token": csrf_token,
                                     "base_code": CASH.base_code,
                                     "title": f"{CASH.title}-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertEqual(account.created_by.username,
                             editor_username)
            self.assertEqual(account.updated_by.username,
                             admin_username)

    def test_l10n(self) -> None:
        """Tests the localization.

        :return: None
        """
        from accounting.models import Account
        detail_uri: str = f"{PREFIX}/{CASH.code}"
        update_uri: str = f"{PREFIX}/{CASH.code}/update"
        account: Account
        response: httpx.Response

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertEqual(account.title_l10n, CASH.title)
            self.assertEqual(account.l10n, [])

        set_locale(self.__app, self.__client, self.__csrf_token, "zh_Hant")

        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": CASH.base_code,
                                            "title": f"{CASH.title}-zh_Hant"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertEqual(account.title_l10n, CASH.title)
            self.assertEqual({(x.locale, x.title) for x in account.l10n},
                             {("zh_Hant", f"{CASH.title}-zh_Hant")})

        set_locale(self.__app, self.__client, self.__csrf_token, "en")

        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": CASH.base_code,
                                            "title": f"{CASH.title}-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertEqual(account.title_l10n, f"{CASH.title}-2")
            self.assertEqual({(x.locale, x.title) for x in account.l10n},
                             {("zh_Hant", f"{CASH.title}-zh_Hant")})

        set_locale(self.__app, self.__client, self.__csrf_token, "zh_Hant")

        response = self.__client.post(update_uri,
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": CASH.base_code,
                                            "title": f"{CASH.title}-zh_Hant-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.__app.app_context():
            account = Account.find_by_code(CASH.code)
            self.assertEqual(account.title_l10n, f"{CASH.title}-2")
            self.assertEqual({(x.locale, x.title) for x in account.l10n},
                             {("zh_Hant", f"{CASH.title}-zh_Hant-2")})

    def test_delete(self) -> None:
        """Tests to delete a currency.

        :return: None.
        """
        from accounting.models import Account
        detail_uri: str = f"{PREFIX}/{PETTY.code}"
        delete_uri: str = f"{PREFIX}/{PETTY.code}/delete"
        list_uri: str = PREFIX
        response: httpx.Response

        response = self.__client.post(f"{PREFIX}/store",
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": PETTY.base_code,
                                            "title": PETTY.title})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        add_journal_entry(self.__client,
                          form={"csrf_token": self.__csrf_token,
                                "next": self.__encoded_next_uri,
                                "date": dt.date.today().isoformat(),
                                "currency-1-code": "USD",
                                "currency-1-credit-1-account_code": BANK.code,
                                "currency-1-credit-1-amount": "20"})

        with self.__app.app_context():
            self.assertEqual({x.code for x in Account.query.all()},
                             {CASH.code, PETTY.code, BANK.code})

        # Cannot delete the cash account
        response = self.__client.post(f"{PREFIX}/{CASH.code}/delete",
                                      data={"csrf_token": self.__csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{CASH.code}")

        # Cannot delete the account that is in use
        response = self.__client.post(f"{PREFIX}/{BANK.code}/delete",
                                      data={"csrf_token": self.__csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{BANK.code}")

        # Success
        response = self.__client.get(detail_uri)
        self.assertEqual(response.status_code, 200)
        response = self.__client.post(delete_uri,
                                      data={"csrf_token": self.__csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], list_uri)

        with self.__app.app_context():
            self.assertEqual({x.code for x in Account.query.all()},
                             {CASH.code, BANK.code})

        response = self.__client.get(detail_uri)
        self.assertEqual(response.status_code, 404)
        response = self.__client.post(delete_uri,
                                      data={"csrf_token": self.__csrf_token})
        self.assertEqual(response.status_code, 404)

    def test_change_base_code(self) -> None:
        """Tests to change the base code of an account.

        :return: None.
        """
        from accounting.models import Account
        response: httpx.Response

        for i in range(2, 6):
            response = self.__client.post(f"{PREFIX}/store",
                                          data={"csrf_token": self.__csrf_token,
                                                "base_code": "1111",
                                                "title": "Title"})
            self.assertEqual(response.status_code, 302)
            self.assertEqual(response.headers["Location"],
                             f"{PREFIX}/1111-00{i}")

        with self.__app.app_context():
            account_1: Account = Account.find_by_code("1111-001")
            id_1: int = account_1.id
            account_2: Account = Account.find_by_code("1111-002")
            id_2: int = account_2.id
            account_3: Account = Account.find_by_code("1111-003")
            id_3: int = account_3.id
            account_4: Account = Account.find_by_code("1111-004")
            id_4: int = account_4.id
            account_5: Account = Account.find_by_code("1111-005")
            id_5: int = account_5.id
            account_1.no = 3
            account_2.no = 5
            account_3.no = 8
            account_4.base_code = "1112"
            account_4.no = 2
            account_5.base_code = "1112"
            account_5.no = 6
            db.session.commit()

        response = self.__client.post(f"{PREFIX}/1111-005/update",
                                      data={"csrf_token": self.__csrf_token,
                                            "base_code": "1112",
                                            "title": "Title"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/1112-003")

        with self.__app.app_context():
            self.assertEqual(db.session.get(Account, id_1).no, 1)
            self.assertEqual(db.session.get(Account, id_2).no, 3)
            self.assertEqual(db.session.get(Account, id_3).no, 2)
            self.assertEqual(db.session.get(Account, id_4).no, 1)
            self.assertEqual(db.session.get(Account, id_5).no, 2)

    def test_reorder(self) -> None:
        """Tests to reorder the accounts under a same base account.

        :return: None.
        """
        from accounting.models import Account
        response: httpx.Response

        for i in range(2, 6):
            response = self.__client.post(f"{PREFIX}/store",
                                          data={"csrf_token": self.__csrf_token,
                                                "base_code": "1111",
                                                "title": "Title"})
            self.assertEqual(response.status_code, 302)
            self.assertEqual(response.headers["Location"],
                             f"{PREFIX}/1111-00{i}")

        # Normal reorder
        with self.__app.app_context():
            id_1: int = Account.find_by_code("1111-001").id
            id_2: int = Account.find_by_code("1111-002").id
            id_3: int = Account.find_by_code("1111-003").id
            id_4: int = Account.find_by_code("1111-004").id
            id_5: int = Account.find_by_code("1111-005").id

        response = self.__client.post(f"{PREFIX}/bases/1111",
                                      data={"csrf_token": self.__csrf_token,
                                            "next": self.__encoded_next_uri,
                                            f"{id_1}-no": "4",
                                            f"{id_2}-no": "1",
                                            f"{id_3}-no": "5",
                                            f"{id_4}-no": "2",
                                            f"{id_5}-no": "3"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], NEXT_URI)

        with self.__app.app_context():
            self.assertEqual(db.session.get(Account, id_1).code, "1111-004")
            self.assertEqual(db.session.get(Account, id_2).code, "1111-001")
            self.assertEqual(db.session.get(Account, id_3).code, "1111-005")
            self.assertEqual(db.session.get(Account, id_4).code, "1111-002")
            self.assertEqual(db.session.get(Account, id_5).code, "1111-003")

        # Malformed orders
        with self.__app.app_context():
            db.session.get(Account, id_1).no = 3
            db.session.get(Account, id_2).no = 4
            db.session.get(Account, id_3).no = 6
            db.session.get(Account, id_4).no = 8
            db.session.get(Account, id_5).no = 9
            db.session.commit()

        response = self.__client.post(f"{PREFIX}/bases/1111",
                                      data={"csrf_token": self.__csrf_token,
                                            "next": self.__encoded_next_uri,
                                            f"{id_2}-no": "3a",
                                            f"{id_3}-no": "5",
                                            f"{id_4}-no": "2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], NEXT_URI)

        with self.__app.app_context():
            self.assertEqual(db.session.get(Account, id_1).code, "1111-003")
            self.assertEqual(db.session.get(Account, id_2).code, "1111-004")
            self.assertEqual(db.session.get(Account, id_3).code, "1111-002")
            self.assertEqual(db.session.get(Account, id_4).code, "1111-001")
            self.assertEqual(db.session.get(Account, id_5).code, "1111-005")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/3

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the independent utilities.

"""
import unittest
from urllib.parse import quote_plus

import httpx
from flask import Flask, request

from accounting.utils.next_uri import append_next, inherit_next, or_next, \
    encode_next, decode_next
from accounting.utils.pagination import Pagination, DEFAULT_PAGE_SIZE
from accounting.utils.query import parse_query_keywords
from testlib import TEST_SERVER, create_test_app, get_csrf_token, NEXT_URI


class NextUriTestCase(unittest.TestCase):
    """The test case for the next URI utilities."""
    TARGET: str = "/target"

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""

    def test_next_uri(self) -> None:
        """Tests the next URI utilities with the next URI.

        :return: None.
        """
        def test_next_uri_view() -> str:
            """The test view with the next URI."""
            current_uri: str = request.full_path if request.query_string \
                else request.path
            with self.__app.app_context():
                encoded_current: str = encode_next(current_uri)
            self.assertEqual(append_next(self.TARGET),
                             f"{self.TARGET}?next={encoded_current}")
            encoded_next_uri: str = request.form["next"] \
                if request.method == "POST" else request.args["next"]
            self.assertEqual(inherit_next(self.TARGET),
                             f"{self.TARGET}?next={encoded_next_uri}")
            with self.__app.app_context():
                next_uri: str = decode_next(encoded_next_uri)
            self.assertEqual(or_next(self.TARGET), next_uri)
            return ""

        self.__app.add_url_rule("/test-next", view_func=test_next_uri_view,
                                methods=["GET", "POST"])
        client: httpx.Client = httpx.Client(app=self.__app,
                                            base_url=TEST_SERVER)
        client.headers["Referer"] = TEST_SERVER
        csrf_token: str = get_csrf_token(client)
        response: httpx.Response

        with self.__app.app_context():
            encoded_uri: str = encode_next(NEXT_URI)
        response = client.get(f"/test-next?next={encoded_uri}&q=abc&page-no=4")
        self.assertEqual(response.status_code, 200)
        response = client.post("/test-next", data={"csrf_token": csrf_token,
                                                   "next": encoded_uri,
                                                   "name": "viewer"})
        self.assertEqual(response.status_code, 200)

    def test_no_next_uri(self) -> None:
        """Tests the next URI utilities without the next URI.

        :return: None.
        """
        def test_no_next_uri_view() -> str:
            """The test view without the next URI."""
            self.assertEqual(inherit_next(self.TARGET), self.TARGET)
            self.assertEqual(or_next(self.TARGET), self.TARGET)
            return ""

        self.__app.add_url_rule("/test-no-next",
                                view_func=test_no_next_uri_view,
                                methods=["GET", "POST"])
        client: httpx.Client = httpx.Client(app=self.__app,
                                            base_url=TEST_SERVER)
        client.headers["Referer"] = TEST_SERVER
        csrf_token: str = get_csrf_token(client)
        response: httpx.Response

        response = client.get("/test-no-next?q=abc&page-no=4")
        self.assertEqual(response.status_code, 200)
        response = client.post("/test-no-next", data={"csrf_token": csrf_token,
                                                      "name": "viewer"})
        self.assertEqual(response.status_code, 200)

    def test_invalid(self) -> None:
        """Tests the next URI utilities without an invalid next URI.

        :return: None.
        """
        def test_invalid_next_uri_view() -> str:
            """The test view without the next URI."""
            self.assertEqual(inherit_next(self.TARGET), self.TARGET)
            self.assertEqual(or_next(self.TARGET), self.TARGET)
            return ""

        self.__app.add_url_rule("/test-invalid-next",
                                view_func=test_invalid_next_uri_view,
                                methods=["GET", "POST"])
        client: httpx.Client = httpx.Client(app=self.__app,
                                            base_url=TEST_SERVER)
        client.headers["Referer"] = TEST_SERVER
        csrf_token: str = get_csrf_token(client)
        next_uri: str
        expected1: str
        expected2: str
        response: httpx.Response

        # A foreign URI
        next_uri = "https://example.com"
        response = client.get(f"/test-invalid-next?next={quote_plus(next_uri)}")
        self.assertEqual(response.status_code, 200)
        response = client.post("/test-invalid-next",
                               data={"csrf_token": csrf_token,
                                     "next": next_uri})
        self.assertEqual(response.status_code, 200)


class QueryKeywordParserTestCase(unittest.TestCase):
    """The test case for the query keyword parser."""

    def test_default(self) -> None:
        """Tests the query keyword parser.

        :return: None.
        """
        self.assertEqual(parse_query_keywords("coffee"), ["coffee"])
        self.assertEqual(parse_query_keywords("coffee tea"), ["coffee", "tea"])
        self.assertEqual(parse_query_keywords("\"coffee\" \"tea cake\""),
                         ["coffee", "tea cake"])
        self.assertEqual(parse_query_keywords("\"coffee tea\" cheese "
                                              "\"cake candy\" sugar"),
                         ["coffee tea", "cheese", "cake candy", "sugar"])

    def test_malformed(self) -> None:
        """Tests the malformed query.

        :return: None.
        """
        self.assertEqual(parse_query_keywords("coffee \"tea cake"),
                         ["coffee", "tea cake"])
        self.assertEqual(parse_query_keywords("coffee te\"a ca\"ke"),
                         ["coffee", "te\"a", "ca\"ke"])
        self.assertEqual(parse_query_keywords("coffee\" tea cake\""),
                         ["coffee\"", "tea", "cake\""])

    def test_empty(self) -> None:
        """Tests the empty query.

        :return: None.
        """
        self.assertEqual(parse_query_keywords(None), [])
        self.assertEqual(parse_query_keywords(""), [])


class PaginationTestCase(unittest.TestCase):
    """The test case for pagination."""

    class Params:
        """The testing pagination parameters."""

        def __init__(self, items: list[int], is_reversed: bool | None,
                     result: list[int], is_paged: bool):
            """Constructs the expected pagination.

            :param items: All the items in the list.
            :param is_reversed: Whether the default page is the last page.
            :param result: The expected items on the page.
            :param is_paged: Whether we need pagination.
            """
            self.items: list[int] = items
            """All the items in the list."""
            self.is_reversed: bool | None = is_reversed
            """Whether the default page is the last page."""
            self.result: list[int] = result
            """The expected items on the page."""
            self.is_paged: bool = is_paged
            """Whether we need pagination."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""
        self.__params: PaginationTestCase.Params \
            = self.Params([], None, [], True)
        """The testing pagination parameters."""

        @self.__app.get("/test-pagination")
        def test_pagination_view() -> str:
            """The test view with the pagination."""
            pagination: Pagination
            if self.__params.is_reversed is not None:
                pagination = Pagination[int](
                    self.__params.items, is_reversed=self.__params.is_reversed)
            else:
                pagination = Pagination[int](self.__params.items)
            self.assertEqual(pagination.is_paged, self.__params.is_paged)
            self.assertEqual(pagination.list, self.__params.result)
            return ""

        self.__client: httpx.Client = httpx.Client(app=self.__app,
                                                   base_url=TEST_SERVER)
        """The user client."""
        self.__client.headers["Referer"] = TEST_SERVER

    def __test_success(self, query: str, items: range,
                       result: range, is_paged: bool = True,
                       is_reversed: bool | None = None) -> None:
        """Tests the pagination.

        :param query: The query string.
        :param items: The original items.
        :param result: The expected page content.
        :param is_paged: Whether we need pagination.
        :param is_reversed: Whether the list is reversed.
        :return: None.
        """
        target: str = "/test-pagination"
        if query != "":
            target = f"{target}?{query}"
        self.__params = self.Params(list(items), is_reversed,
                                    list(result), is_paged)
        response: httpx.Response = self.__client.get(target)
        self.assertEqual(response.status_code, 200)

    def __test_malformed(self, query: str, items: range, redirect_to: str,
                         is_reversed: bool | None = None) -> None:
        """Tests the pagination.

        :param query: The query string.
        :param items: The original items.
        :param redirect_to: The expected target query of the redirection.
        :param is_reversed: Whether the list is reversed.
        :return: None.
        """
        target: str = "/test-pagination"
        self.__params = self.Params(list(items), is_reversed, [], True)
        response: httpx.Response = self.__client.get(f"{target}?{query}")
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{target}?{redirect_to}")

    def test_default(self) -> None:
        """Tests the default pagination.

        :return: None.
        """
        # The default first page
        self.__test_success("", range(1, 687), range(1, 11))
        # Some page in the middle
        self.__test_success("page-no=37", range(1, 687), range(361, 371))
        # The last page
        self.__test_success("page-no=69", range(1, 687), range(681, 687))

    def test_page_size(self) -> None:
        """Tests the pagination with a different page size.

        :return: None.
        """
        # The default page with a different page size
        self.__test_success("page-size=15", range(1, 687), range(1, 16))
        # Some page with a different page size
        self.__test_success("page-no=37&page-size=15", range(1, 687),
                            range(541, 556))
        # The last page with a different page size.
        self.__test_success("page-no=46&page-size=15", range(1, 687),
                            range(676, 687))

    def test_not_need(self) -> None:
        """Tests that the data does not need pagination.

        :return: None.
        """
        # Empty list
        self.__test_success("", range(0, 0), range(0, 0), is_paged=False)
        # A list that fits in one page
        self.__test_success("", range(1, 4), range(1, 4), is_paged=False)
        # A large page size that fits in everything
        self.__test_success("page-size=1000", range(1, 687), range(1, 687),
                            is_paged=False)

    def test_reversed(self) -> None:
        """Tests the default page on a reversed list.

        :return: None.
        """
        # The default page
        self.__test_success("", range(1, 687), range(681, 687),
                            is_reversed=True)
        # The default page with a different page size
        self.__test_success("page-size=15", range(1, 687), range(676, 687),
                            is_reversed=True)

    def test_last_page(self) -> None:
        """Tests the calculation of the items on the last page.

        :return: None.
        """
        # The last page that fits in one page
        self.__test_success("page-no=69", range(1, 691), range(681, 691))
        # A danging item in the last page
        self.__test_success("page-no=70", range(1, 692), range(691, 692))

    def test_malformed(self) -> None:
        """Tests the malformed pagination parameters.

        :return: None.
        """
        page_37: str = "q=word&page-no=37&next=%2F"
        page_size_15_default: str = "q=word&page-size=15&next=%2F"

        # A malformed page size
        self.__test_malformed("q=word&page-size=100a&page-no=37&next=%2F",
                              range(1, 691), page_37)
        # A default page size
        self.__test_malformed(f"q=word&page-size={DEFAULT_PAGE_SIZE}"
                              "&page-no=37&next=%2F",
                              range(1, 691), page_37)
        # An invalid page size
        self.__test_malformed("q=word&page-size=0&page-no=37&next=%2F",
                              range(1, 691), page_37)
        # A malformed page number
        self.__test_malformed("q=word&page-size=15&page-no=37a&next=%2F",
                              range(1, 691), page_size_15_default)
        # A default page number
        self.__test_malformed("q=word&page-size=15&page-no=1&next=%2F",
                              range(1, 691), page_size_15_default)
        # A default page number, on a reversed list
        self.__test_malformed("q=word&page-size=15&page-no=46&next=%2F",
                              range(1, 691), page_size_15_default,
                              is_reversed=True)
        # A page number beyond the last page
        self.__test_malformed("q=word&page-size=15&page-no=100&next=%2F",
                              range(1, 691),
                              "q=word&page-size=15&page-no=46&next=%2F")
        # A page number beyond the last page, on a reversed list
        self.__test_malformed("q=word&page-size=15&page-no=100&next=%2F",
                              range(1, 691),
                              page_size_15_default, is_reversed=True)
        # A page number before the first page
        self.__test_malformed("q=word&page-size=15&page-no=0&next=%2F",
                              range(1, 691),
                              page_size_15_default)
        # A page number before the first page, on a reversed list
        self.__test_malformed("q=word&page-size=15&page-no=0&next=%2F",
                              range(1, 691),
                              "q=word&page-size=15&page-no=1&next=%2F",
                              is_reversed=True)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The date period.

This file is largely taken from the NanoParma ERP project, first written in
2021/9/16 by imacat (imacat@nanoparma.com).

"""
import datetime as dt
from typing import Self

from .description import get_desc
from .month_end import month_end
from .specification import get_spec


class Period:
    """A date period."""

    def __init__(self, start: dt.date | None, end: dt.date | None):
        """Constructs a new date period.

        :param start: The start date, or None from the very beginning.
        :param end: The end date, or None till no end.
        """
        self.start: dt.date | None = start
        """The start of the period."""
        self.end: dt.date | None = end
        """The end of the period."""
        self.is_default: bool = False
        """Whether this is the default period."""
        self.is_this_month: bool = False
        """Whether the period is this month."""
        self.is_last_month: bool = False
        """Whether the period is last month."""
        self.is_since_last_month: bool = False
        """Whether the period is since last month."""
        self.is_this_year: bool = False
        """Whether the period is this year."""
        self.is_last_year: bool = False
        """Whether the period is last year."""
        self.is_today: bool = False
        """Whether the period is today."""
        self.is_yesterday: bool = False
        """Whether the period is yesterday."""
        self.is_all: bool = start is None and end is None
        """Whether the period is all time."""
        self.spec: str = ""
        """The period specification."""
        self.desc: str = ""
        """The text description."""
        self.is_a_month: bool = False
        """Whether the period is a whole month."""
        self.is_type_month: bool = False
        """Whether the period is for the month chooser."""
        self.is_a_year: bool = False
        """Whether the period is a whole year."""
        self.is_a_day: bool = False
        """Whether the period is a single day."""
        self._set_properties()

    def _set_properties(self) -> None:
        """Sets the following properties.

        * self.spec
        * self.desc
        * self.is_a_month
        * self.is_type_month
        * self.is_a_year
        * self.is_a_day

        Override this method to set the properties in the subclasses, to skip
        the calculation.

        :return: None.
        """
        self.spec = get_spec(self.start, self.end)
        self.desc = get_desc(self.start, self.end)
        if self.start is None or self.end is None:
            return
        self.is_a_month = self.start.day == 1 \
            and self.end == month_end(self.start)
        self.is_type_month = self.is_a_month
        self.is_a_year = self.start == dt.date(self.start.year, 1, 1) \
            and self.end == dt.date(self.start.year, 12, 31)
        self.is_a_day = self.start == self.end

    def is_year(self, year: int) -> bool:
        """Returns whether the period is the specific year period.

        :param year: The year.
        :return: True if the period is the year period, or False otherwise.
        """
        if not self.is_a_year:
            return False
        return self.start.year == year

    @property
    def is_type_arbitrary(self) -> bool:
        """Returns whether this period is an arbitrary period.

        :return: True if this is an arbitrary period, or False otherwise.
        """
        return not self.is_type_month and not self.is_a_year \
            and not self.is_a_day

    @property
    def before(self) -> Self | None:
        """Returns the period before this period.

        :return: The period before this period.
        """
        if self.start is None:
            return None
        return Period(None, self.start - dt.timedelta(days=1))



# The Mia! Accounting Demonstration Website.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/13

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The common library for the Mia! Accounting demonstration website.

"""
from __future__ import annotations

import datetime as dt
from abc import ABC, abstractmethod
from decimal import Decimal
from secrets import randbelow
from typing import Any

import sqlalchemy as sa
from flask import Flask

from . import db
from .auth import User


class Accounts:
    """The shortcuts to the common accounts."""
    CASH: str = "1111-001"
    BANK: str = "1113-001"
    RECEIVABLE: str = "1141-001"
    MACHINERY: str = "1441-001"
    PAYABLE: str = "2141-001"
    SERVICE: str = "4611-001"
    RENT_EXPENSE: str = "6252-001"
    MEAL: str = "6272-001"


class JournalEntryLineItemData:
    """The journal entry line item data."""

    def __init__(self, account: str, description: str | None, amount: str,
                 original_line_item: JournalEntryLineItemData | None = None):
        """Constructs the journal entry line item data.

        :param account: The account code.
        :param description: The description.
        :param amount: The amount.
        :param original_line_item: The original journal entry line item.
        """
        self.journal_entry: JournalEntryData | None = None
        """The journal entry data."""
        self.id: int = -1
        """The journal entry line item ID."""
        self.no: int = -1
        """The line item number under the journal entry and debit or credit."""
        self.original_line_item: JournalEntryLineItemData | None \
            = original_line_item
        """The original journal entry line item."""
        self.account: str = account
        """The account code."""
        self.description: str | None = description
        """The description."""
        self.amount: Decimal = Decimal(amount)
        """The amount."""

    def form(self, prefix: str, debit_credit: str, index: int,
             is_update: bool) -> dict[str, str]:
        """Returns the line item as form data.

        :param prefix: The prefix of the form fields.
        :param debit_credit: Either "debit" or "credit".
        :param index: The line item index.
        :param is_update: True for an update operation, or False otherwise
        :return: The form data.
        """
        prefix = f"{prefix}-{debit_credit}-{index}"
        form: dict[str, str] = {f"{prefix}-account_code": self.account,
                                f"{prefix}-description": self.description,
                                f"{prefix}-amount": str(self.amount)}
        if is_update and self.id != -1:
            form[f"{prefix}-id"] = str(self.id)
        form[f"{prefix}-no"] = str(index) if self.no == -1 else str(self.no)
        if self.original_line_item is not None:
            assert self.original_line_item.id != -1
            form[f"{prefix}-original_line_item_id"] \
                = str(self.original_line_item.id)
        return form


class JournalEntryCurrencyData:
    """The journal entry currency data."""

    def __init__(self, currency: str, debit: list[JournalEntryLineItemData],
                 credit: list[JournalEntryLineItemData]):
        """Constructs the journal entry currency data.

        :param currency: The currency code.
        :param debit: The debit line items.
        :param credit: The credit line items.
        """
        self.code: str = currency
        """The currency code."""
        self.debit: list[JournalEntryLineItemData] = debit
        """The debit line items."""
        self.credit: list[JournalEntryLineItemData] = credit
        """The credit line items."""

    def form(self, index: int, is_update: bool) -> dict[str, str]:
        """Returns the currency as form data.

        :param index: The currency index.
        :param is_update: True for an update operation, or False otherwise
        :return: The form data.
        """
        prefix: str = f"currency-{index}"
        form: dict[str, str] = {f"{prefix}-code": self.code}
        for i in range(len(self.debit)):
            form.update(self.debit[i].form(prefix, "debit", i + 1, is_update))
        for i in range(len(self.credit)):
            form.update(self.credit[i].form(prefix, "credit", i + 1,
                                            is_update))
        return form


class JournalEntryData:
    """The journal entry data."""

    def __init__(self, days: int, currencies: list[JournalEntryCurrencyData]):
        """Constructs a journal entry.

        :param days: The number of days before today.
        :param currencies: The journal entry currency data.
        """
        self.id: int = -1
        """The journal entry ID."""
        self.days: int = days
        """The number of days before today."""
        self.currencies: list[JournalEntryCurrencyData] = currencies
        """The journal entry currency data."""
        self.note: str | None = None
        """The note."""
        for currency in self.currencies:
            for line_item in currency.debit:
                line_item.journal_entry = self
            for line_item in currency.credit:
                line_item.journal_entry = self

    def new_form(self, csrf_token: str, encoded_next_uri: str) \
            -> dict[str, str]:
        """Returns the journal entry as a creation form.

        :param csrf_token: The CSRF token.
        :param encoded_next_uri: The encoded next URI.
        :return: The journal entry as a creation form.
        """
        return self.__form(csrf_token, encoded_next_uri, is_update=False)

    def update_form(self, csrf_token: str, encoded_next_uri: str) \
            -> dict[str, str]:
        """Returns the journal entry as an update form.

        :param csrf_token: The CSRF token.
        :param encoded_next_uri: The encoded next URI.
        :return: The journal entry as an update form.
        """
        return self.__form(csrf_token, encoded_next_uri, is_update=True)

    def __form(self, csrf_token: str, encoded_next_uri: str,
               is_update: bool = False) -> dict[str, str]:
        """Returns the journal entry as a form.

        :param csrf_token: The CSRF token.
        :param encoded_next_uri: The encoded next URI.
        :param is_update: True for an update operation, or False otherwise
        :return: The journal entry as a form.
        """
        date: dt.date = dt.date.today() - dt.timedelta(days=self.days)
        form: dict[str, str] = {"csrf_token": csrf_token,
                                "next": encoded_next_uri,
                                "date": date.isoformat()}
        for i in range(len(self.currencies)):
            form.update(self.currencies[i].form(i + 1, is_update))
        if self.note is not None:
            form["note"] = self.note
        return form


class BaseTestData(ABC):
    """The base test data."""

    def __init__(self, app: Flask, username: str):
        """Constructs the test data.

        :param app: The Flask application.
        :param username: The username.
        """
        self._app: Flask = app
        """The Flask application."""
        with self._app.app_context():
            current_user: User | None = User.query\
                .filter(User.username == username).first()
            assert current_user is not None
            self.__current_user_id: int = current_user.id
            """The current user ID."""
            self.__journal_entries: list[dict[str, Any]] = []
            """The data of the journal entries."""
            self.__line_items: list[dict[str, Any]] = []
            """The data of the journal entry line items."""
            self._init_data()

    @abstractmethod
    def _init_data(self) -> None:
        """Initializes the test data.

        :return: None
        """

    def populate(self) -> None:
        """Populates the data into the database.

        :return: None
        """
        from accounting.models import JournalEntry, JournalEntryLineItem
        with self._app.app_context():
            db.session.execute(sa.insert(JournalEntry), self.__journal_entries)
            db.session.execute(sa.insert(JournalEntryLineItem),
                               self.__line_items)
            db.session.commit()

    @staticmethod
    def _couple(description: str, amount: str, debit: str, credit: str) \
            -> tuple[JournalEntryLineItemData, JournalEntryLineItemData]:
        """Returns a couple of debit-credit line items.

        :param description: The description.
        :param amount: The amount.
        :param debit: The debit account code.
        :param credit: The credit account code.
        :return: The debit line item and credit line item.
        """
        return JournalEntryLineItemData(debit, description, amount),\
            JournalEntryLineItemData(credit, description, amount)

    def _add_journal_entry(self, journal_entry_data: JournalEntryData) -> None:
        """Adds a journal entry.

        :param journal_entry_data: The journal entry data.
        :return: None.
        """
        from accounting.models import Account
        existing_j_id: set[int] = {x["id"] for x in self.__journal_entries}
        existing_l_id: set[int] = {x["id"] for x in self.__line_items}
        journal_entry_data.id = self.__new_id(existing_j_id)
        date: dt.date \
            = dt.date.today() - dt.timedelta(days=journal_entry_data.days)
        self.__journal_entries.append(
            {"id": journal_entry_data.id,
             "date": date,
             "no": self.__next_j_no(date),
             "note": journal_entry_data.note,
             "created_by_id": self.__current_user_id,
             "updated_by_id": self.__current_user_id})
        debit_no: int = 0
        credit_no: int = 0
        for currency in journal_entry_data.currencies:
            for line_item in currency.debit:
                account: Account | None \
                    = Account.find_by_code(line_item.account)
                assert account is not None
                debit_no = debit_no + 1
                line_item.id = self.__new_id(existing_l_id)
                data: dict[str, Any] \
                    = {"id": line_item.id,
                       "journal_entry_id": journal_entry_data.id,
                       "is_debit": True,
                       "no": debit_no,
                       "account_id": account.id,
                       "currency_code": currency.code,
                       "description": line_item.description,
                       "amount": line_item.amount}
                if line_item.original_line_item is not None:
                    data["original_line_item_id"] \
                        = line_item.original_line_item.id
                self.__line_items.append(data)
            for line_item in currency.credit:
                account: Account | None \
                    = Account.find_by_code(line_item.account)
                assert account is not None
                credit_no = credit_no + 1
                line_item.id = self.__new_id(existing_l_id)
                data: dict[str, Any] \
                    = {"id": line_item.id,
                       "journal_entry_id": journal_entry_data.id,
                       "is_debit": False,
                       "no": credit_no,
                       "account_id": account.id,
                       "currency_code": currency.code,
                       "description": line_item.description,
                       "amount": line_item.amount}
                if line_item.original_line_item is not None:
                    data["original_line_item_id"] \
                        = line_item.original_line_item.id
                self.__line_items.append(data)

    @staticmethod
    def __new_id(existing_id: set[int]) -> int:
        """Generates and returns a new random unique ID.

        :param existing_id: The existing ID.
        :return: The newly-generated random unique ID.
        """
        while True:
            obj_id: int = 100000000 + randbelow(900000000)
            if obj_id not in existing_id:
                existing_id.add(obj_id)
                return obj_id

    def __next_j_no(self, date: dt.date) -> int:
        """Returns the next journal entry number in a day.

        :param date: The journal entry date.
        :return: The next journal entry number.
        """
        existing: set[int] = {x["no"] for x in self.__journal_entries
                              if x["date"] == date}
        return 1 if len(existing) == 0 else max(existing) + 1

    def _add_simple_journal_entry(
            self, days: int, currency: str, description: str, amount: str,
            debit: str, credit: str) \
            -> tuple[JournalEntryLineItemData, JournalEntryLineItemData]:
        """Adds a simple journal entry.

        :param days: The number of days before today.
        :param currency: The currency code.
        :param description: The description.
        :param amount: The amount.
        :param debit: The debit account code.
        :param credit: The credit account code.
        :return: The debit line item and credit line item.
        """
        debit_item, credit_item = self._couple(
            description, amount, debit, credit)
        self._add_journal_entry(JournalEntryData(
            days, [JournalEntryCurrencyData(
                currency, [debit_item], [credit_item])]))
        return debit_item, credit_item



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/8

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The base report.

"""
from abc import ABC, abstractmethod

from flask import Response


class BaseReport(ABC):
    """The base report class."""

    @abstractmethod
    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """

    @abstractmethod
    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The period chooser.

This file is largely taken from the NanoParma ERP project, first written in
2021/9/16 by imacat (imacat@nanoparma.com).

"""
import datetime as dt
from collections.abc import Callable

from accounting.models import JournalEntry
from .period import Period
from .shortcuts import ThisMonth, LastMonth, SinceLastMonth, ThisYear, \
    LastYear, Today, Yesterday, AllTime, TemplatePeriod, YearPeriod


class PeriodChooser:
    """The period chooser."""

    def __init__(self, get_url: Callable[[Period], str]):
        """Constructs a period chooser.

        :param get_url: The callback to return the URL of the current report in
            a period.
        """
        self.__get_url: Callable[[Period], str] = get_url
        """The callback to return the URL of the current report in a period."""

        # Shortcut periods
        self.this_month_url: str = get_url(ThisMonth())
        """The URL for this month."""
        self.last_month_url: str = get_url(LastMonth())
        """The URL for last month."""
        self.since_last_month_url: str = get_url(SinceLastMonth())
        """The URL since last mint."""
        self.this_year_url: str = get_url(ThisYear())
        """The URL for this year."""
        self.last_year_url: str = get_url(LastYear())
        """The URL for last year."""
        self.today_url: str = get_url(Today())
        """The URL for today."""
        self.yesterday_url: str = get_url(Yesterday())
        """The URL for yesterday."""
        self.all_url: str = get_url(AllTime())
        """The URL for all period."""
        self.url_template: str = get_url(TemplatePeriod())
        """The URL template."""

        first: JournalEntry | None \
            = JournalEntry.query.order_by(JournalEntry.date).first()
        start: dt.date | None = None if first is None else first.date

        # Attributes
        self.data_start: dt.date | None = start
        """The start of the data."""
        self.has_data: bool = start is not None
        """Whether there is any data."""
        self.has_last_month: bool = False
        """Where there is data in last month."""
        self.has_last_year: bool = False
        """Whether there is data in last year."""
        self.has_yesterday: bool = False
        """Whether there is data in yesterday."""
        self.available_years: list[int] = []
        """The available years."""

        if self.has_data:
            today: dt.date = dt.date.today()
            self.has_last_month = start < dt.date(today.year, today.month, 1)
            self.has_last_year = start.year < today.year
            self.has_yesterday = start < today
            if start.year < today.year - 1:
                self.available_years \
                    = reversed(range(start.year, today.year - 1))

    def year_url(self, year: int) -> str:
        """Returns the period URL of a year.

        :param year: The year
        :return: The period URL of the year.
        """
        return self.__get_url(YearPeriod(year))



# The Mia! Accounting Demonstration Website.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/27

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The authentication for the Mia! Accounting demonstration website.

"""
from collections.abc import Callable

from flask import Blueprint, render_template, Flask, redirect, url_for, \
    session, request, g, Response, abort
from sqlalchemy.orm import Mapped, mapped_column

from . import db

bp: Blueprint = Blueprint("auth", __name__, url_prefix="/")
"""The authentication blueprint."""


class User(db.Model):
    """A user."""
    __tablename__ = "users"
    """The table name."""
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    """The ID."""
    username: Mapped[str] = mapped_column(unique=True)
    """The username."""

    def __str__(self) -> str:
        """Returns the string representation of the user.

        :return: The string representation of the user.
        """
        return self.username


@bp.get("login", endpoint="login-form")
def show_login_form() -> str | Response:
    """Shows the login form.

    :return: The login form.
    """
    if "user" in session:
        return redirect(url_for("accounting-report.default"))
    return render_template("login.html")


@bp.post("login", endpoint="login")
def login() -> redirect:
    """Logs in the user.

    :return: The redirection to the home page.
    """
    from accounting.utils.next_uri import inherit_next, or_next
    if request.form.get("username") not in {"viewer", "editor", "admin",
                                            "nobody"}:
        return redirect(inherit_next(url_for("auth.login")))
    session["user"] = request.form.get("username")
    return redirect(or_next(url_for("accounting-report.default")))


@bp.post("logout", endpoint="logout")
def logout() -> redirect:
    """Logs out the user.

    :return: The redirection to the home page.
    """
    if "user" in session:
        del session["user"]
    return redirect(url_for("home.home"))


def current_user() -> User | None:
    """Returns the current user.

    :return: The current user, or None if the user did not log in.
    """
    if not hasattr(g, "user"):
        if "user" not in session:
            g.user = None
        else:
            g.user = User.query.filter(
                User.username == session["user"]).first()
    return g.user


def admin_required(view: Callable) -> Callable:
    """The view decorator to require the user to be an administrator.

    :param view: The view.
    :return: The decorated view.
    """

    def decorated_view(*args, **kwargs):
        """The decorated view that tests against a permission rule.

        :param args: The arguments of the view.
        :param kwargs: The keyword arguments of the view.
        :return: The response of the view.
        :raise Forbidden: When the user is denied.
        """
        from accounting.utils.next_uri import append_next
        if "user" not in session:
            return redirect(append_next(url_for("auth.login")))
        if session["user"] != "admin":
            abort(403)
        return view(*args, **kwargs)

    return decorated_view


def init_app(app: Flask) -> None:
    """Initialize the localization.

    :param app: The Flask application.
    :return: None.
    """
    app.register_blueprint(bp)
    app.jinja_env.globals["current_user"] = current_user



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/27

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The description editor.

"""
import re
from typing import Literal

import sqlalchemy as sa

from accounting import db
from accounting.models import Account, JournalEntryLineItem
from accounting.utils.options import options, Recurring


class DescriptionAccount:
    """An account for a description tag."""

    def __init__(self, account: Account, freq: int):
        """Constructs an account for a description tag.

        :param account: The account.
        :param freq: The frequency of the tag with the account.
        """
        self.__account: Account = account
        """The account."""
        self.id: int = account.id
        """The account ID."""
        self.code: str = account.code
        """The account code."""
        self.is_need_offset: bool = account.is_need_offset
        """Whether the journal entry line items of this account need offset."""
        self.freq: int = freq
        """The frequency of the tag with the account."""

    def __str__(self) -> str:
        """Returns the string representation of the account.

        :return: The string representation of the account.
        """
        return str(self.__account)

    @property
    def title(self) -> str:
        """Returns the account title.

        :return: The account title.
        """
        return self.__account.title

    def add_freq(self, freq: int) -> None:
        """Adds the frequency of an account.

        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        self.freq = self.freq + freq


class DescriptionTag:
    """A description tag."""

    def __init__(self, name: str):
        """Constructs a description tag.

        :param name: The tag name.
        """
        self.name: str = name
        """The tag name."""
        self.__account_dict: dict[int, DescriptionAccount] = {}
        """The accounts that come with the tag, in the order of their
        frequency."""
        self.freq: int = 0
        """The frequency of the tag."""

    def __str__(self) -> str:
        """Returns the string representation of the tag.

        :return: The string representation of the tag.
        """
        return self.name

    def add_account(self, account: Account, freq: int):
        """Adds an account.

        :param account: The associated account.
        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        self.__account_dict[account.id] = DescriptionAccount(account, freq)
        self.freq = self.freq + freq

    @property
    def accounts(self) -> list[DescriptionAccount]:
        """Returns the accounts by the order of their frequencies.

        :return: The accounts by the order of their frequencies.
        """
        return sorted(self.__account_dict.values(), key=lambda x: -x.freq)

    @property
    def account_codes(self) -> list[str]:
        """Returns the account codes by the order of their frequencies.

        :return: The account codes by the order of their frequencies.
        """
        return [x.code for x in self.accounts]


class DescriptionType:
    """A description type"""

    def __init__(self, type_id: Literal["general", "travel", "bus"]):
        """Constructs a description type.

        :param type_id: The type ID, either "general", "travel", or "bus".
        """
        self.id: Literal["general", "travel", "bus"] = type_id
        """The type ID."""
        self.__tag_dict: dict[str, DescriptionTag] = {}
        """A dictionary from the tag name to their corresponding tag."""

    def add_tag(self, name: str, account: Account, freq: int) -> None:
        """Adds a tag.

        :param name: The tag name.
        :param account: The associated account.
        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        if name not in self.__tag_dict:
            self.__tag_dict[name] = DescriptionTag(name)
        self.__tag_dict[name].add_account(account, freq)

    @property
    def tags(self) -> list[DescriptionTag]:
        """Returns the tags by the order of their frequencies.

        :return: The tags by the order of their frequencies.
        """
        return sorted(self.__tag_dict.values(), key=lambda x: -x.freq)


class DescriptionRecurring:
    """A recurring transaction."""

    def __init__(self, name: str, account: Account, description_template: str):
        """Constructs a recurring transaction.

        :param name: The name.
        :param description_template: The description template.
        :param account: The account.
        """
        self.name: str = name
        """The name."""
        self.account: DescriptionAccount = DescriptionAccount(account, 0)
        """The account."""
        self.description_template: str = description_template
        """The description template."""

    @property
    def account_codes(self) -> list[str]:
        """Returns the account codes by the order of their frequencies.

        :return: The account codes by the order of their frequencies.
        """
        return [self.account.code]


class DescriptionDebitCredit:
    """The description on debit or credit."""

    def __init__(self, debit_credit: Literal["debit", "credit"]):
        """Constructs the description on debit or credit.

        :param debit_credit: Either "debit" or "credit".
        """
        self.debit_credit: Literal["debit", "credit"] = debit_credit
        """Either debit or credit."""
        self.general: DescriptionType = DescriptionType("general")
        """The general tags."""
        self.travel: DescriptionType = DescriptionType("travel")
        """The travel tags."""
        self.bus: DescriptionType = DescriptionType("bus")
        """The bus tags."""
        self.__type_dict: dict[Literal["general", "travel", "bus"],
                               DescriptionType] \
            = {x.id: x for x in {self.general, self.travel, self.bus}}
        """A dictionary from the type ID to the corresponding tags."""
        self.recurring: list[DescriptionRecurring] = []
        """The recurring transactions."""

    def add_tag(self, tag_type: Literal["general", "travel", "bus"],
                name: str, account: Account, freq: int) -> None:
        """Adds a tag.

        :param tag_type: The tag type, either "general", "travel", or "bus".
        :param name: The name.
        :param account: The associated account.
        :param freq: The frequency of the tag name with the account.
        :return: None.
        """
        self.__type_dict[tag_type].add_tag(name, account, freq)

    @property
    def accounts(self) -> list[DescriptionAccount]:
        """Returns the suggested accounts of all tags in the description editor
        in debit or credit, in their frequency order.

        :return: The suggested accounts of all tags, in their frequency order.
        """
        accounts: dict[int, DescriptionAccount] = {}
        freq: dict[int, int] = {}
        for tag_type in self.__type_dict.values():
            for tag in tag_type.tags:
                for account in tag.accounts:
                    accounts[account.id] = account
                    if account.id not in freq:
                        freq[account.id] = 0
                    freq[account.id] \
                        = freq[account.id] + account.freq
        for recurring in self.recurring:
            accounts[recurring.account.id] = recurring.account
            if recurring.account.id not in freq:
                freq[recurring.account.id] = 0
        return [accounts[y] for y in sorted(freq.keys(),
                                            key=lambda x: -freq[x])]


class DescriptionEditor:
    """The description editor."""

    def __init__(self):
        """Constructs the description editor."""
        self.debit: DescriptionDebitCredit = DescriptionDebitCredit("debit")
        """The debit tags."""
        self.credit: DescriptionDebitCredit = DescriptionDebitCredit("credit")
        """The credit tags."""
        self.__init_tags()
        self.__init_recurring()

    def __init_tags(self):
        """Initializes the tags.

        :return: None.
        """
        debit_credit: sa.Label = sa.case(
            (JournalEntryLineItem.is_debit, "debit"),
            else_="credit").label("debit_credit")
        tag_type: sa.Label = sa.case(
            (JournalEntryLineItem.description.like("_%—_%—_%→_%"), "bus"),
            (sa.or_(JournalEntryLineItem.description.like("_%—_%→_%"),
                    JournalEntryLineItem.description.like("_%—_%↔_%")),
             "travel"),
            else_="general").label("tag_type")
        tag: sa.Label = get_prefix(JournalEntryLineItem.description, "—")\
            .label("tag")
        select: sa.Select = sa.Select(debit_credit, tag_type, tag,
                                      JournalEntryLineItem.account_id,
                                      sa.func.count().label("freq"))\
            .filter(JournalEntryLineItem.description.is_not(None),
                    JournalEntryLineItem.description.like("_%—_%"),
                    JournalEntryLineItem.original_line_item_id.is_(None))\
            .group_by(debit_credit, tag_type, tag,
                      JournalEntryLineItem.account_id)
        result: list[sa.Row] = db.session.execute(select).all()
        accounts: dict[int, Account] \
            = {x.id: x for x in Account.query
               .filter(Account.id.in_({x.account_id for x in result})).all()}
        debit_credit_dict: dict[Literal["debit", "credit"],
                                DescriptionDebitCredit] \
            = {x.debit_credit: x for x in {self.debit, self.credit}}
        for row in result:
            debit_credit_dict[row.debit_credit].add_tag(
                row.tag_type, row.tag, accounts[row.account_id], row.freq)

    def __init_recurring(self) -> None:
        """Initializes the recurring transactions.

        :return: None.
        """
        recurring: Recurring = options.recurring
        accounts: dict[str, Account] \
            = self.__get_accounts(recurring.codes)
        self.debit.recurring \
            = [DescriptionRecurring(x.name, accounts[x.account_code],
                                    x.description_template)
               for x in recurring.expenses]
        self.credit.recurring \
            = [DescriptionRecurring(x.name, accounts[x.account_code],
                                    x.description_template)
               for x in recurring.incomes]

    @staticmethod
    def __get_accounts(codes: set[str]) -> dict[str, Account]:
        """Finds and returns the accounts by codes.

        :param codes: The account codes.
        :return: The account.
        """
        if len(codes) == 0:
            return {}

        def get_condition(code0: str) -> sa.BinaryExpression:
            m: re.Match = re.match(r"^(\d{4})-(\d{3})$", code0)
            assert m is not None,\
                f"Malformed account code \"{code0}\" for regular transactions."
            return sa.and_(Account.base_code == m.group(1),
                           Account.no == int(m.group(2)))

        conditions: list[sa.BinaryExpression] \
            = [get_condition(x) for x in codes]
        accounts: dict[str, Account] \
            = {x.code: x for x in
               Account.query.filter(sa.or_(*conditions)).all()}
        for code in codes:
            assert code in accounts,\
                f"Unknown account \"{code}\" for regular transactions."
        return accounts


def get_prefix(string: str | sa.Column, separator: str | sa.Column) \
        -> sa.Function:
    """Returns the SQL function to find the prefix of a string.

    :param string: The string.
    :param separator: The separator.
    :return: The position of the substring, starting from 1.
    """
    return sa.func.substr(string, 0, get_position(string, separator))


def get_position(string: str | sa.Column, substring: str | sa.Column) \
        -> sa.Function:
    """Returns the SQL function to find the position of a substring.

    :param string: The string.
    :param substring: The substring.
    :return: The position of the substring, starting from 1.
    """
    if db.engine.name == "postgresql":
        return sa.func.strpos(string, substring)
    return sa.func.instr(string, substring)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/22

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The views for the option management.

"""
from urllib.parse import parse_qsl, urlencode

from flask import Blueprint, render_template, redirect, session, request, \
    flash, url_for
from werkzeug.datastructures import ImmutableMultiDict

from accounting.locale import lazy_gettext
from accounting.utils.cast import s
from accounting.utils.flash_errors import flash_form_errors
from accounting.utils.next_uri import inherit_next
from accounting.utils.options import options
from accounting.utils.permission import has_permission, can_admin
from .forms import OptionForm

bp: Blueprint = Blueprint("option", __name__)
"""The view blueprint for the currency management."""


@bp.get("", endpoint="detail")
@has_permission(can_admin)
def show_options() -> str:
    """Shows the options.

    :return: The options.
    """
    return render_template("accounting/option/detail.html", obj=options)


@bp.get("edit", endpoint="edit")
@has_permission(can_admin)
def show_option_form() -> str:
    """Shows the option form.

    :return: The option form.
    """
    form: OptionForm
    if "form" in session:
        form = OptionForm(ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.validate()
    else:
        form = OptionForm(obj=options)
    return render_template("accounting/option/form.html", form=form)


@bp.post("update", endpoint="update")
@has_permission(can_admin)
def update_options() -> redirect:
    """Updates the options.

    :return: The redirection to the option form.
    """
    form = OptionForm(request.form)
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(url_for("accounting.option.edit")))
    form.populate_obj(options)
    if not options.is_modified:
        flash(s(lazy_gettext("The settings were not modified.")), "success")
        return redirect(inherit_next(url_for("accounting.option.detail")))
    options.commit()
    flash(s(lazy_gettext("The settings are saved successfully.")), "success")
    return redirect(inherit_next(url_for("accounting.option.detail")))



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The line item sub-forms for the journal entry management.

"""
import datetime as dt
from decimal import Decimal

import sqlalchemy as sa
from flask_babel import LazyString
from flask_wtf import FlaskForm
from sqlalchemy.orm import selectinload
from wtforms import StringField, ValidationError, DecimalField, IntegerField
from wtforms.validators import Optional

from accounting import db
from accounting.forms import ACCOUNT_REQUIRED, AccountExists, IsDebitAccount, \
    IsCreditAccount
from accounting.locale import lazy_gettext
from accounting.models import Account, JournalEntry, JournalEntryLineItem
from accounting.template_filters import format_amount
from accounting.utils.random_id import new_id
from accounting.utils.strip_text import strip_text
from accounting.utils.user import get_current_user_pk


class OriginalLineItemExists:
    """The validator to check if the original line item exists."""

    def __call__(self, form: FlaskForm, field: IntegerField) -> None:
        if field.data is None:
            return
        if db.session.get(JournalEntryLineItem, field.data) is None:
            raise ValidationError(lazy_gettext(
                "The original line item does not exist."))


class OriginalLineItemOppositeDebitCredit:
    """The validator to check if the original line item is on the opposite
    debit or credit."""

    def __call__(self, form: FlaskForm, field: IntegerField) -> None:
        if field.data is None:
            return
        original_line_item: JournalEntryLineItem | None \
            = db.session.get(JournalEntryLineItem, field.data)
        if original_line_item is None:
            return
        if isinstance(form, CreditLineItemForm) \
                and original_line_item.is_debit:
            return
        if isinstance(form, DebitLineItemForm) \
                and not original_line_item.is_debit:
            return
        raise ValidationError(lazy_gettext(
            "The original line item is on the same debit or credit."))


class OriginalLineItemNeedOffset:
    """The validator to check if the original line item needs offset."""

    def __call__(self, form: FlaskForm, field: IntegerField) -> None:
        if field.data is None:
            return
        original_line_item: JournalEntryLineItem | None \
            = db.session.get(JournalEntryLineItem, field.data)
        if original_line_item is None:
            return
        if not original_line_item.account.is_need_offset:
            raise ValidationError(lazy_gettext(
                "The original line item does not need offset."))


class OriginalLineItemNotOffset:
    """The validator to check if the original line item is not itself an
    offset item."""

    def __call__(self, form: FlaskForm, field: IntegerField) -> None:
        if field.data is None:
            return
        original_line_item: JournalEntryLineItem | None \
            = db.session.get(JournalEntryLineItem, field.data)
        if original_line_item is None:
            return
        if original_line_item.original_line_item_id is not None:
            raise ValidationError(lazy_gettext(
                "The original line item cannot be an offset item."))


class SameAccountAsOriginalLineItem:
    """The validator to check if the account is the same as the
    original line item."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, LineItemForm)
        if field.data is None or form.original_line_item_id.data is None:
            return
        original_line_item: JournalEntryLineItem | None \
            = db.session.get(JournalEntryLineItem,
                             form.original_line_item_id.data)
        if original_line_item is None:
            return
        if field.data != original_line_item.account_code:
            raise ValidationError(lazy_gettext(
                "The account must be the same as the original line item."))


class KeepAccountWhenHavingOffset:
    """The validator to check if the account is the same when having offset."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, LineItemForm)
        if field.data is None or form.id.data is None:
            return
        line_item: JournalEntryLineItem | None \
            = db.session.get(JournalEntryLineItem, form.id.data)
        if line_item is None or len(line_item.offsets) == 0:
            return
        if field.data != line_item.account_code:
            raise ValidationError(lazy_gettext(
                "The account must not be changed when there is offset."))


class NotStartPayableFromDebit:
    """The validator to check that a payable line item does not start from
    debit."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, DebitLineItemForm)
        if field.data is None \
                or field.data[0] != "2" \
                or form.original_line_item_id.data is not None:
            return
        account: Account | None = Account.find_by_code(field.data)
        if account is not None and account.is_need_offset:
            raise ValidationError(lazy_gettext(
                "A payable line item cannot start from debit."))


class NotStartReceivableFromCredit:
    """The validator to check that a receivable line item does not start
    from credit."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, CreditLineItemForm)
        if field.data is None \
                or field.data[0] != "1" \
                or form.original_line_item_id.data is not None:
            return
        account: Account | None = Account.find_by_code(field.data)
        if account is not None and account.is_need_offset:
            raise ValidationError(lazy_gettext(
                "A receivable line item cannot start from credit."))


class PositiveAmount:
    """The validator to check if the amount is positive."""

    def __call__(self, form: FlaskForm, field: DecimalField) -> None:
        if field.data is None:
            return
        if field.data <= 0:
            raise ValidationError(lazy_gettext(
                "Please fill in a positive amount."))


class NotExceedingOriginalLineItemNetBalance:
    """The validator to check if the amount exceeds the net balance of the
    original line item."""

    def __call__(self, form: FlaskForm, field: DecimalField) -> None:
        assert isinstance(form, LineItemForm)
        if field.data is None or form.original_line_item_id.data is None:
            return
        original_line_item: JournalEntryLineItem | None \
            = db.session.get(JournalEntryLineItem,
                             form.original_line_item_id.data)
        if original_line_item is None:
            return
        is_debit: bool = isinstance(form, DebitLineItemForm)
        existing_line_item_id: set[int] = set()
        if form.journal_entry_form.obj is not None:
            existing_line_item_id \
                = {x.id for x in form.journal_entry_form.obj.line_items}
        offset_total_func: sa.Function = sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit == is_debit,
             JournalEntryLineItem.amount),
            else_=-JournalEntryLineItem.amount))
        offset_total_but_form: Decimal | None = db.session.scalar(
            sa.select(offset_total_func)
            .filter(JournalEntryLineItem.original_line_item_id
                    == original_line_item.id,
                    JournalEntryLineItem.id.not_in(existing_line_item_id)))
        if offset_total_but_form is None:
            offset_total_but_form = Decimal("0")
        offset_total_on_form: Decimal = sum(
            [x.amount.data for x in form.journal_entry_form.line_items
             if x.original_line_item_id.data == original_line_item.id
             and x.amount != field and x.amount.data is not None])
        net_balance: Decimal = original_line_item.amount \
            - offset_total_but_form - offset_total_on_form
        if field.data > net_balance:
            raise ValidationError(lazy_gettext(
                "The amount must not exceed the net balance %(balance)s of the"
                " original line item.", balance=format_amount(net_balance)))


class NotLessThanOffsetTotal:
    """The validator to check if the amount is less than the offset total."""

    def __call__(self, form: FlaskForm, field: DecimalField) -> None:
        assert isinstance(form, LineItemForm)
        if field.data is None or form.id.data is None:
            return
        is_debit: bool = isinstance(form, DebitLineItemForm)
        select_offset_total: sa.Select = sa.select(sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit != is_debit,
             JournalEntryLineItem.amount),
            else_=-JournalEntryLineItem.amount)))\
            .filter(JournalEntryLineItem.original_line_item_id == form.id.data)
        offset_total: Decimal | None = db.session.scalar(select_offset_total)
        if offset_total is not None and field.data < offset_total:
            raise ValidationError(lazy_gettext(
                "The amount must not be less than the offset total %(total)s.",
                total=format_amount(offset_total)))


class LineItemForm(FlaskForm):
    """The base form to create or edit a line item."""
    id = IntegerField()
    """The existing line item ID."""
    no = IntegerField()
    """The order in the currency."""
    original_line_item_id = IntegerField()
    """The Id of the original line item."""
    account_code = StringField()
    """The account code."""
    description = StringField()
    """The description."""
    amount = DecimalField()
    """The amount."""

    def __init__(self, *args, **kwargs):
        """Constructs a base line item form.

        :param args: The arguments.
        :param kwargs: The keyword arguments.
        """
        super().__init__(*args, **kwargs)
        from .journal_entry import JournalEntryForm
        self.journal_entry_form: JournalEntryForm | None = None
        """The source journal entry form."""

    @property
    def account_title(self) -> str:
        """Returns the title of the account.

        :return: The title of the account.
        """
        if self.account_code.data is None:
            return ""
        account: Account | None = Account.find_by_code(self.account_code.data)
        if account is None:
            return ""
        return account.title

    @property
    def account_text(self) -> str:
        """Returns the text representation of the account.

        :return: The text representation of the account.
        """
        if self.account_code.data is None:
            return ""
        account: Account | None = Account.find_by_code(self.account_code.data)
        if account is None:
            return ""
        return str(account)

    @property
    def __original_line_item(self) -> JournalEntryLineItem | None:
        """Returns the original line item.

        :return: The original line item.
        """
        if not hasattr(self, "____original_line_item"):
            def get_line_item() -> JournalEntryLineItem | None:
                if self.original_line_item_id.data is None:
                    return None
                return db.session.get(JournalEntryLineItem,
                                      self.original_line_item_id.data)
            setattr(self, "____original_line_item", get_line_item())
        return getattr(self, "____original_line_item")

    @property
    def original_line_item_date(self) -> dt.date | None:
        """Returns the text representation of the original line item.

        :return: The text representation of the original line item.
        """
        return None if self.__original_line_item is None \
            else self.__original_line_item.journal_entry.date

    @property
    def original_line_item_text(self) -> str | None:
        """Returns the text representation of the original line item.

        :return: The text representation of the original line item.
        """
        return None if self.__original_line_item is None \
            else str(self.__original_line_item)

    @property
    def is_need_offset(self) -> bool:
        """Returns whether the line item needs offset.

        :return: True if the line item needs offset, or False otherwise.
        """
        if self.account_code.data is None:
            return False
        if self.account_code.data[0] == "1":
            if isinstance(self, CreditLineItemForm):
                return False
        elif self.account_code.data[0] == "2":
            if isinstance(self, DebitLineItemForm):
                return False
        else:
            return False
        account: Account | None = Account.find_by_code(self.account_code.data)
        return account is not None and account.is_need_offset

    @property
    def offsets(self) -> list[JournalEntryLineItem]:
        """Returns the offsets.

        :return: The offsets.
        """
        if not hasattr(self, "__offsets"):
            def get_offsets() -> list[JournalEntryLineItem]:
                if not self.is_need_offset or self.id.data is None:
                    return []
                return JournalEntryLineItem.query.join(JournalEntry)\
                    .filter(JournalEntryLineItem.original_line_item_id
                            == self.id.data)\
                    .order_by(JournalEntry.date, JournalEntry.no,
                              JournalEntryLineItem.no)\
                    .options(selectinload(JournalEntryLineItem.journal_entry),
                             selectinload(JournalEntryLineItem.account)).all()
            setattr(self, "__offsets", get_offsets())
        return getattr(self, "__offsets")

    @property
    def offset_total(self) -> Decimal | None:
        """Returns the total amount of the offsets.

        :return: The total amount of the offsets.
        """
        if not hasattr(self, "__offset_total"):
            def get_offset_total():
                if not self.is_need_offset or self.id.data is None:
                    return None
                is_debit: bool = isinstance(self, DebitLineItemForm)
                return sum([x.amount if x.is_debit != is_debit else -x.amount
                            for x in self.offsets])
            setattr(self, "__offset_total", get_offset_total())
        return getattr(self, "__offset_total")

    @property
    def net_balance(self) -> Decimal | None:
        """Returns the net balance.

        :return: The net balance.
        """
        if not self.is_need_offset or self.id.data is None \
                or self.amount.data is None:
            return None
        return self.amount.data - self.offset_total

    @property
    def all_errors(self) -> list[str | LazyString]:
        """Returns all the errors of the form.

        :return: All the errors of the form.
        """
        all_errors: list[str | LazyString] = []
        for key in self.errors:
            if key != "csrf_token":
                all_errors.extend(self.errors[key])
        return all_errors


class DebitLineItemForm(LineItemForm):
    """The form to create or edit a debit line item."""
    id = IntegerField()
    """The existing line item ID."""
    no = IntegerField()
    """The order in the currency."""
    original_line_item_id = IntegerField(
        validators=[Optional(),
                    OriginalLineItemExists(),
                    OriginalLineItemOppositeDebitCredit(),
                    OriginalLineItemNeedOffset(),
                    OriginalLineItemNotOffset()])
    """The ID of the original line item."""
    account_code = StringField(
        filters=[strip_text],
        validators=[
            ACCOUNT_REQUIRED,
            AccountExists(),
            IsDebitAccount(lazy_gettext(
                "This account is not for debit line items.")),
            SameAccountAsOriginalLineItem(),
            KeepAccountWhenHavingOffset(),
            NotStartPayableFromDebit()])
    """The account code."""
    description = StringField(filters=[strip_text])
    """The description."""
    amount = DecimalField(
        validators=[PositiveAmount(),
                    NotExceedingOriginalLineItemNetBalance(),
                    NotLessThanOffsetTotal()])
    """The amount."""

    def populate_obj(self, obj: JournalEntryLineItem) -> None:
        """Populates the form data into a line item object.

        :param obj: The line item object.
        :return: None.
        """
        is_new: bool = obj.id is None
        if is_new:
            obj.id = new_id(JournalEntryLineItem)
        obj.original_line_item_id = self.original_line_item_id.data
        obj.account_id = Account.find_by_code(self.account_code.data).id
        obj.description = self.description.data
        obj.is_debit = True
        obj.amount = self.amount.data
        if is_new:
            current_user_pk: int = get_current_user_pk()
            obj.created_by_id = current_user_pk
            obj.updated_by_id = current_user_pk


class CreditLineItemForm(LineItemForm):
    """The form to create or edit a credit line item."""
    id = IntegerField()
    """The existing line item ID."""
    no = IntegerField()
    """The order in the currency."""
    original_line_item_id = IntegerField(
        validators=[Optional(),
                    OriginalLineItemExists(),
                    OriginalLineItemOppositeDebitCredit(),
                    OriginalLineItemNeedOffset(),
                    OriginalLineItemNotOffset()])
    """The ID of the original line item."""
    account_code = StringField(
        filters=[strip_text],
        validators=[
            ACCOUNT_REQUIRED,
            AccountExists(),
            IsCreditAccount(lazy_gettext(
                "This account is not for credit line items.")),
            SameAccountAsOriginalLineItem(),
            KeepAccountWhenHavingOffset(),
            NotStartReceivableFromCredit()])
    """The account code."""
    description = StringField(filters=[strip_text])
    """The description."""
    amount = DecimalField(
        validators=[PositiveAmount(),
                    NotExceedingOriginalLineItemNetBalance(),
                    NotLessThanOffsetTotal()])
    """The amount."""

    def populate_obj(self, obj: JournalEntryLineItem) -> None:
        """Populates the form data into a line item object.

        :param obj: The line item object.
        :return: None.
        """
        is_new: bool = obj.id is None
        if is_new:
            obj.id = new_id(JournalEntryLineItem)
        obj.original_line_item_id = self.original_line_item_id.data
        obj.account_id = Account.find_by_code(self.account_code.data).id
        obj.description = self.description.data
        obj.is_debit = False
        obj.amount = self.amount.data
        if is_new:
            current_user_pk: int = get_current_user_pk()
            obj.created_by_id = current_user_pk
            obj.updated_by_id = current_user_pk



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The period specification composer.

"""
import datetime as dt


def get_spec(start: dt.date | None, end: dt.date | None) -> str:
    """Returns the period specification.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The period specification.
    """
    if start is None and end is None:
        return "-"
    if end is None:
        return __get_since_spec(start)
    if start is None:
        return __get_until_spec(end)
    try:
        return __get_year_spec(start, end)
    except ValueError:
        pass
    try:
        return __get_month_spec(start, end)
    except ValueError:
        pass
    return __get_day_spec(start, end)


def __get_since_spec(start: dt.date) -> str:
    """Returns the period specification without the end day.

    :param start: The start of the period.
    :return: The period specification without the end day
    """
    if start.month == 1 and start.day == 1:
        return start.strftime("%Y-")
    if start.day == 1:
        return start.strftime("%Y-%m-")
    return start.strftime("%Y-%m-%d-")


def __get_until_spec(end: dt.date) -> str:
    """Returns the period specification without the start day.

    :param end: The end of the period.
    :return: The period specification without the start day
    """
    if end.month == 12 and end.day == 31:
        return end.strftime("-%Y")
    if (end + dt.timedelta(days=1)).day == 1:
        return end.strftime("-%Y-%m")
    return end.strftime("-%Y-%m-%d")


def __get_year_spec(start: dt.date, end: dt.date) -> str:
    """Returns the period specification as a year range.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The period specification as a year range.
    :raise ValueError: The period is not a year range.
    """
    if start.month != 1 or start.day != 1 \
            or end.month != 12 or end.day != 31:
        raise ValueError
    start_spec: str = start.strftime("%Y")
    if start.year == end.year:
        return start_spec
    end_spec: str = end.strftime("%Y")
    return f"{start_spec}-{end_spec}"


def __get_month_spec(start: dt.date, end: dt.date) -> str:
    """Returns the period specification as a month range.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The period specification as a month range.
    :raise ValueError: The period is not a month range.
    """
    if start.day != 1 or (end + dt.timedelta(days=1)).day != 1:
        raise ValueError
    start_spec: str = start.strftime("%Y-%m")
    if start.year == end.year and start.month == end.month:
        return start_spec
    end_spec: str = end.strftime("%Y-%m")
    return f"{start_spec}-{end_spec}"


def __get_day_spec(start: dt.date, end: dt.date) -> str:
    """Returns the period specification as a day range.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The period specification as a day range.
    :raise ValueError: The period is a month or year range.
    """
    start_spec: str = start.strftime("%Y-%m-%d")
    if start == end:
        return start_spec
    end_spec: str = end.strftime("%Y-%m-%d")
    return f"{start_spec}-{end_spec}"



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The user utilities.

This module should not import any other module from the application.

"""
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Type

import sqlalchemy as sa
from flask import g, Response
from flask_sqlalchemy.model import Model

T = TypeVar("T", bound=Model)
"""The user data model data type."""


class UserUtilityInterface(Generic[T], ABC):
    """The interface for the user utilities."""

    @abstractmethod
    def can_view(self) -> bool:
        """Returns whether the currently logged-in user can view the accounting
        data.

        :return: True if the currently logged-in user can view the accounting
            data, or False otherwise.
        """

    @abstractmethod
    def can_edit(self) -> bool:
        """Returns whether the currently logged-in user can edit the accounting
        data.

        :return: True if the currently logged-in user can edit the accounting
            data, or False otherwise.
        """

    @abstractmethod
    def can_admin(self) -> bool:
        """Returns whether the currently logged-in user can administrate the
        accounting settings.

        :return: True if the currently logged-in user can administrate the
            accounting settings, or False otherwise.
        """

    @abstractmethod
    def unauthorized(self) -> Response | None:
        """Returns the response to require the user to log in.

        This may be a redirection to the login page, or an HTTP 401
        Unauthorized response for HTTP Authentication.  If this returns None,
        an HTTP 403 Forbidden response is return to the user.

        :return: The response to require the user to log in.
        """

    @property
    @abstractmethod
    def cls(self) -> Type[T]:
        """Returns the class of the user data model.

        :return: The class of the user data model.
        """

    @property
    @abstractmethod
    def pk_column(self) -> sa.Column:
        """Returns the primary key column.

        :return: The primary key column.
        """

    @property
    @abstractmethod
    def current_user(self) -> T | None:
        """Returns the currently logged-in user.

        :return: The currently logged-in user, or None if the user has not
            logged in
        """

    @abstractmethod
    def get_by_username(self, username: str) -> T | None:
        """Returns the user by her username.

        :return: The user by her username, or None if the user was not found.
        """

    @abstractmethod
    def get_pk(self, user: T) -> int:
        """Returns the primary key of the user, as an integer.

        :return: The primary key of the user, as an integer.
        """


__user_utils: UserUtilityInterface
"""The user utilities."""
user_cls: Type[Model] = Model
"""The user class."""
user_pk_column: sa.Column = sa.Column(sa.Integer)
"""The primary key column of the user class."""


def init_user_utils(utils: UserUtilityInterface) -> None:
    """Initializes the user utilities.

    :param utils: The user utilities.
    :return: None.
    """
    global __user_utils, user_cls, user_pk_column
    __user_utils = utils
    user_cls = utils.cls
    user_pk_column = utils.pk_column


def get_current_user_pk() -> int:
    """Returns the primary key value of the currently logged-in user.

    :return: The primary key value of the currently logged-in user.
    """
    return __user_utils.get_pk(get_current_user())


def has_user(username: str) -> bool:
    """Returns whether a user by the username exists.

    :param username: The username.
    :return: True if the user by the username exists, or False otherwise.
    """
    return __user_utils.get_by_username(username) is not None


def get_user_pk(username: str) -> int:
    """Returns the primary key value of the user by the username.

    :param username: The username.
    :return: The primary key value of the user by the username.
    """
    return __user_utils.get_pk(__user_utils.get_by_username(username))


def get_current_user() -> user_cls | None:
    """Returns the currently logged-in user.  The result is cached in the
    current request.

    :return: The currently logged-in user.
    """
    if not hasattr(g, "_accounting_user"):
        setattr(g, "_accounting_user", __user_utils.current_user)
    return getattr(g, "_accounting_user")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/5

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The option link.

"""


class OptionLink:
    """An option link."""

    def __init__(self, title: str, url: str, is_active: bool,
                 fa_icon: str | None = None):
        """Constructs an option link.

        :param title: The title.
        :param url: The URL.
        :param is_active: True if active, or False otherwise
        :param fa_icon: The font-awesome icon, if any.
        """
        self.title: str = title
        """The title."""
        self.url: str = url
        """The URL."""
        self.is_active: bool = is_active
        """True if active, or False otherwise."""
        self.fa_icon: str | None = fa_icon
        """The font-awesome icon, if any."""



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/6

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The forms for the currency management.

"""
from flask_wtf import FlaskForm
from wtforms import StringField, ValidationError
from wtforms.validators import DataRequired, Regexp, NoneOf

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import Currency
from accounting.utils.strip_text import strip_text
from accounting.utils.user import get_current_user_pk


class CodeUnique:
    """The validator to check if the code is unique."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        assert isinstance(form, CurrencyForm)
        if field.data == "":
            return
        if form.obj_code is not None and form.obj_code == field.data:
            return
        if db.session.get(Currency, field.data) is not None:
            raise ValidationError(lazy_gettext(
                "Code conflicts with another currency."))


class CurrencyForm(FlaskForm):
    """The form to create or edit a currency."""
    CODE_BLOCKLIST: list[str] = ["create", "store", "exists-code"]
    """The reserved codes that are not available."""
    code = StringField(
        filters=[strip_text],
        validators=[DataRequired(lazy_gettext("Please fill in the code.")),
                    Regexp(r"^[A-Z]{3}$",
                           message=lazy_gettext(
                               "Code can only be composed of 3 upper-cased"
                               " letters.")),
                    NoneOf(CODE_BLOCKLIST, message=lazy_gettext(
                        "This code is not available.")),
                    CodeUnique()])
    """The code.  It may not conflict with another currency."""
    name = StringField(
        filters=[strip_text],
        validators=[DataRequired(lazy_gettext("Please fill in the name."))])
    """The name."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj_code: str | None = None
        """The current code of the currency, or None when adding a new
        currency."""

    def populate_obj(self, obj: Currency) -> None:
        """Populates the form data into a currency object.

        :param obj: The currency object.
        :return: None.
        """
        is_new: bool = obj.code is None
        obj.code = self.code.data
        obj.name = self.name.data
        if is_new:
            current_user_pk: int = get_current_user_pk()
            obj.created_by_id = current_user_pk
            obj.updated_by_id = current_user_pk



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The balance sheet.

"""
from decimal import Decimal

import sqlalchemy as sa
from flask import render_template, Response

from accounting import db
from accounting.locale import gettext
from accounting.models import Currency, BaseAccount, Account, JournalEntry, \
    JournalEntryLineItem
from accounting.report.period import Period, PeriodChooser
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download, \
    period_spec
from accounting.report.utils.option_link import OptionLink
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.urls import ledger_url, balance_sheet_url, \
    income_statement_url


class ReportAccount:
    """An account in the report."""

    def __init__(self, account: Account, amount: Decimal, url: str):
        """Constructs an account in the report.

        :param account: The account.
        :param amount: The amount.
        :param url: The URL to the ledger of the account.
        """
        self.account: Account = account
        """The account."""
        self.amount: Decimal = amount
        """The amount of the account."""
        self.url: str = url
        """The URL to the ledger of the account."""


class Subsection:
    """A subsection."""

    def __init__(self, title: BaseAccount):
        """Constructs a subsection.

        :param title: The title account.
        """
        self.title: BaseAccount = title
        """The title account."""
        self.accounts: list[ReportAccount] = []
        """The accounts in the subsection."""

    @property
    def total(self) -> Decimal:
        """Returns the total of the subsection.

        :return: The total of the subsection.
        """
        return sum([x.amount for x in self.accounts])


class Section:
    """A section."""

    def __init__(self, title: BaseAccount):
        """Constructs a section.

        :param title: The title account.
        """
        self.title: BaseAccount = title
        """The title account."""
        self.subsections: list[Subsection] = []
        """The subsections in the section."""

    @property
    def total(self) -> Decimal:
        """Returns the total of the section.

        :return: The total of the section.
        """
        return sum([x.total for x in self.subsections])


class AccountCollector:
    """The balance sheet account collector."""

    def __init__(self, currency: Currency, period: Period):
        """Constructs the balance sheet account collector.

        :param currency: The currency.
        :param period: The period.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__period: Period = period
        """The period."""
        self.accounts: list[ReportAccount] = self.__query_balances()
        """The balance sheet accounts."""

    def __query_balances(self) -> list[ReportAccount]:
        """Queries and returns the balances.

        :return: The balances.
        """
        sub_conditions: list[sa.BinaryExpression] \
            = [Account.base_code.startswith(x) for x in {"1", "2", "3"}]
        conditions: list[sa.BinaryExpression] \
            = [JournalEntryLineItem.currency_code == self.__currency.code,
               sa.or_(*sub_conditions)]
        if self.__period.end is not None:
            conditions.append(JournalEntry.date <= self.__period.end)
        balance_func: sa.Function = sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit, JournalEntryLineItem.amount),
            else_=-JournalEntryLineItem.amount)).label("balance")
        select_balance: sa.Select \
            = sa.select(Account.id, Account.base_code, Account.no,
                        balance_func)\
            .join(JournalEntry).join(Account)\
            .filter(*conditions)\
            .group_by(Account.id, Account.base_code, Account.no)\
            .having(balance_func != 0)\
            .order_by(Account.base_code, Account.no)
        account_balances: list[sa.Row] \
            = db.session.execute(select_balance).all()
        self.__all_accounts: list[Account] = Account.query\
            .filter(sa.or_(Account.id.in_({x.id for x in account_balances}),
                           Account.base_code == "3351",
                           Account.base_code == "3353")).all()
        """The accounts."""
        account_by_id: dict[int, Account] \
            = {x.id: x for x in self.__all_accounts}
        self.accounts: list[ReportAccount] \
            = [ReportAccount(account=account_by_id[x.id],
                             amount=x.balance,
                             url=ledger_url(self.__currency,
                                            account_by_id[x.id],
                                            self.__period))
               for x in account_balances]
        """The accounts on the balance sheet."""
        self.__add_accumulated()
        self.__add_current_period()
        self.accounts.sort(key=lambda x: (x.account.base_code, x.account.no))
        for balance in self.accounts:
            if not balance.account.base_code.startswith("1"):
                balance.amount = -balance.amount
        return self.accounts

    def __add_accumulated(self) -> None:
        """Adds the accumulated profit or loss to the balances.

        :return: None.
        """
        self.__add_owner_s_equity(Account.ACCUMULATED_CHANGE_CODE,
                                  self.__query_accumulated(),
                                  self.__period)

    def __query_accumulated(self) -> Decimal | None:
        """Queries and returns the accumulated profit or loss.

        :return: The accumulated profit or loss.
        """
        if self.__period.start is None:
            return None
        conditions: list[sa.BinaryExpression] \
            = [JournalEntryLineItem.currency_code == self.__currency.code,
               JournalEntry.date < self.__period.start]
        return self.__query_balance(conditions)

    def __add_current_period(self) -> None:
        """Adds the accumulated profit or loss to the balances.

        :return: None.
        """
        self.__add_owner_s_equity(Account.NET_CHANGE_CODE,
                                  self.__query_current_period(),
                                  self.__period)

    def __query_current_period(self) -> Decimal | None:
        """Queries and returns the net income or loss for current period.

        :return: The net income or loss for current period.
        """
        conditions: list[sa.BinaryExpression] \
            = [JournalEntryLineItem.currency_code == self.__currency.code]
        if self.__period.start is not None:
            conditions.append(JournalEntry.date >= self.__period.start)
        if self.__period.end is not None:
            conditions.append(JournalEntry.date <= self.__period.end)
        return self.__query_balance(conditions)

    @staticmethod
    def __query_balance(conditions: list[sa.BinaryExpression])\
            -> Decimal:
        """Queries the balance.

        :param conditions: The SQL conditions for the balance.
        :return: The balance.
        """
        conditions.extend([sa.not_(Account.base_code.startswith(x))
                           for x in {"1", "2", "3"}])
        balance_func: sa.Function = sa.func.sum(sa.case(
            (JournalEntryLineItem.is_debit, JournalEntryLineItem.amount),
            else_=-JournalEntryLineItem.amount))
        select_balance: sa.Select = sa.select(balance_func)\
            .join(JournalEntry).join(Account).filter(*conditions)
        return db.session.scalar(select_balance)

    def __add_owner_s_equity(self, code: str, amount: Decimal | None,
                             period: Period) -> None:
        """Adds an owner's equity balance.

        :param code: The code of the account to add.
        :param amount: The amount.
        :param period: The period.
        :return: None.
        """
        if amount is None:
            return
        url: str = income_statement_url(self.__currency, period)
        # There is an existing balance.
        account_balance_by_code: dict[str, ReportAccount] \
            = {x.account.code: x for x in self.accounts}
        if code in account_balance_by_code:
            balance: ReportAccount = account_balance_by_code[code]
            balance.amount = balance.amount + amount
            balance.url = url
            return
        # Add a new balance
        account_by_code: dict[str, Account] \
            = {x.code: x for x in self.__all_accounts}
        self.accounts.append(ReportAccount(account=account_by_code[code],
                                           amount=amount,
                                           url=url))


class CSVHalfRow:
    """A half row in the CSV."""

    def __init__(self, title: str | None, amount: Decimal | None):
        """The constructs a half row in the CSV.

        :param title: The title.
        :param amount: The amount.
        """
        self.title: str | None = title
        """The title."""
        self.amount: Decimal | None = amount
        """The amount."""


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self):
        """Constructs a row in the CSV."""
        self.asset_title: str | None = None
        """The title of the asset."""
        self.asset_amount: Decimal | None = None
        """The amount of the asset."""
        self.liability_title: str | None = None
        """The title of the liability."""
        self.liability_amount: Decimal | None = None
        """The amount of the liability."""

    @property
    def values(self) -> list[str | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.asset_title, self.asset_amount,
                self.liability_title, self.liability_amount]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, currency: Currency,
                 period: Period,
                 has_data: bool,
                 assets: Section,
                 liabilities: Section,
                 owner_s_equity: Section):
        """Constructs the HTML page parameters.

        :param currency: The currency.
        :param period: The period.
        :param has_data: True if there is any data, or False otherwise.
        :param assets: The assets.
        :param liabilities: The liabilities.
        :param owner_s_equity: The owner's equity.
        """
        self.currency: Currency = currency
        """The currency."""
        self.period: Period = period
        """The period."""
        self.__has_data: bool = has_data
        """True if there is any data, or False otherwise."""
        self.assets: Section = assets
        """The assets."""
        self.liabilities: Section = liabilities
        """The liabilities."""
        self.owner_s_equity: Section = owner_s_equity
        """The owner's equity."""
        self.period_chooser: PeriodChooser = PeriodChooser(
            lambda x: balance_sheet_url(currency, x))
        """The period chooser."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return self.__has_data

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.BALANCE_SHEET,
                             currency=self.currency,
                             period=self.period)

    @property
    def currency_options(self) -> list[OptionLink]:
        """Returns the currency options.

        :return: The currency options.
        """
        return self._get_currency_options(
            lambda x: balance_sheet_url(x, self.period), self.currency)


class BalanceSheet(BaseReport):
    """The balance sheet."""

    def __init__(self, currency: Currency, period: Period):
        """Constructs a balance sheet.

        :param currency: The currency.
        :param period: The period.
        """
        self.__currency: Currency = currency
        """The currency."""
        self.__period: Period = period
        """The period."""
        self.__has_data: bool
        """True if there is any data, or False otherwise."""
        self.__assets: Section
        """The assets."""
        self.__liabilities: Section
        """The liabilities."""
        self.__owner_s_equity: Section
        """The owner's equity."""
        self.__set_data()

    def __set_data(self) -> None:
        """Queries and sets assets, the liabilities, and the owner's equity
        sections in the balance sheet.

        :return: None.
        """
        balances: list[ReportAccount] = AccountCollector(
            self.__currency, self.__period).accounts

        titles: list[BaseAccount] = BaseAccount.query\
            .filter(BaseAccount.code.in_({"1", "2", "3"})).all()
        subtitles: list[BaseAccount] = BaseAccount.query\
            .filter(BaseAccount.code.in_({x.account.base_code[:2]
                                          for x in balances})).all()

        sections: dict[str, Section] = {x.code: Section(x) for x in titles}
        subsections: dict[str, Subsection] = {x.code: Subsection(x)
                                              for x in subtitles}
        for subsection in subsections.values():
            sections[subsection.title.code[0]].subsections.append(subsection)
        for balance in balances:
            subsections[balance.account.base_code[:2]].accounts.append(balance)

        self.__has_data = len(balances) > 0
        self.__assets = sections["1"]
        self.__liabilities = sections["2"]
        self.__owner_s_equity = sections["3"]

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = "balance-sheet-{currency}-{period}.csv"\
            .format(currency=self.__currency.code,
                    period=period_spec(self.__period))
        return csv_download(filename, self.__get_csv_rows())

    def __get_csv_rows(self) -> list[CSVRow]:
        """Composes and returns the CSV rows.

        :return: The CSV rows.
        """
        asset_rows: list[CSVHalfRow] = self.__section_csv_rows(self.__assets)
        liability_rows: list[CSVHalfRow] = []
        liability_rows.extend(self.__section_csv_rows(self.__liabilities))
        liability_rows.append(CSVHalfRow(gettext("Total"),
                                         self.__liabilities.total))
        liability_rows.append(CSVHalfRow(None, None))
        liability_rows.extend(self.__section_csv_rows(self.__owner_s_equity))
        liability_rows.append(CSVHalfRow(gettext("Total"),
                                         self.__owner_s_equity.total))
        rows: list[CSVRow] = [CSVRow() for _ in
                              range(max(len(asset_rows), len(liability_rows)))]
        for i in range(len(rows)):
            if i < len(asset_rows):
                rows[i].asset_title = asset_rows[i].title
                rows[i].asset_amount = asset_rows[i].amount
            if i < len(liability_rows) and liability_rows[i].title is not None:
                rows[i].liability_title = liability_rows[i].title
                rows[i].liability_amount = liability_rows[i].amount
        total: CSVRow = CSVRow()
        total.asset_title = gettext("Total")
        total.asset_amount = self.__assets.total
        total.liability_title = gettext("Total")
        total.liability_amount \
            = self.__liabilities.total + self.__owner_s_equity.total
        rows.append(total)
        return rows

    @staticmethod
    def __section_csv_rows(section: Section) -> list[CSVHalfRow]:
        """Gathers the CSV rows for a section.

        :param section: The section.
        :return: The CSV rows for the section.
        """
        rows: list[CSVHalfRow] \
            = [CSVHalfRow(section.title.title, None)]
        for subsection in section.subsections:
            rows.append(CSVHalfRow(f" {subsection.title.title}", None))
            for account in subsection.accounts:
                rows.append(CSVHalfRow(f"  {str(account.account)}",
                                       account.amount))
        return rows

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        params: PageParams = PageParams(currency=self.__currency,
                                        period=self.__period,
                                        has_data=self.__has_data,
                                        assets=self.__assets,
                                        liabilities=self.__liabilities,
                                        owner_s_equity=self.__owner_s_equity)
        return render_template("accounting/report/balance-sheet.html",
                               report=params)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/30

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The random ID utility for the data models.

This module should not import any other module from the application.

"""
from secrets import randbelow
from typing import Type

from accounting import db


def new_id(cls: Type[db.Model]):
    """Generates and returns a new, unused random ID for the data model.

    :param cls: The data model.
    :return: The newly-generated, unused random ID.
    """
    while True:
        obj_id: int = 100000000 + randbelow(900000000)
        if db.session.get(cls, obj_id) is None:
            return obj_id



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The template filters for the reports.

"""
from decimal import Decimal

from accounting.template_filters import format_amount as core_format_amount


def format_amount(value: Decimal | None) -> str | None:
    """Formats an amount for the report.

    :param value: The amount.
    :return: The formatted amount text.
    """
    if value is None:
        return ""
    is_negative: bool = value < 0
    formatted: str = core_format_amount(abs(value))
    if is_negative:
        formatted = f"({formatted})"
    return formatted



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/19

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The operators for different journal entry types.

"""
from abc import ABC, abstractmethod
from typing import Type

from flask import render_template, request, abort
from flask_wtf import FlaskForm

from accounting.journal_entry.forms import JournalEntryForm, \
    CashReceiptJournalEntryForm, CashDisbursementJournalEntryForm, \
    TransferJournalEntryForm
from accounting.journal_entry.forms.line_item import LineItemForm
from accounting.models import JournalEntry
from accounting.template_globals import default_currency_code
from accounting.utils.journal_entry_types import JournalEntryType


class JournalEntryOperator(ABC):
    """The base journal entry operator."""
    CHECK_ORDER: int = -1
    """The order when checking the journal entry operator."""

    @property
    @abstractmethod
    def form(self) -> Type[JournalEntryForm]:
        """Returns the form class.

        :return: The form class.
        """

    @abstractmethod
    def render_create_template(self, form: FlaskForm) -> str:
        """Renders the template for the form to create a journal entry.

        :param form: The journal entry form.
        :return: the form to create a journal entry.
        """

    @abstractmethod
    def render_detail_template(self, journal_entry: JournalEntry) -> str:
        """Renders the template for the detail page.

        :param journal_entry: The journal entry.
        :return: the detail page.
        """

    @abstractmethod
    def render_edit_template(self, journal_entry: JournalEntry,
                             form: FlaskForm) -> str:
        """Renders the template for the form to edit a journal entry.

        :param journal_entry: The journal entry.
        :param form: The form.
        :return: the form to edit a journal entry.
        """

    @abstractmethod
    def is_my_type(self, journal_entry: JournalEntry) -> bool:
        """Checks and returns whether the journal entry belongs to the type.

        :param journal_entry: The journal entry.
        :return: True if the journal entry belongs to the type, or False
            otherwise.
        """

    @property
    def _line_item_template(self) -> str:
        """Renders and returns the template for the line item sub-form.

        :return: The template for the line item sub-form.
        """
        return render_template(
            "accounting/journal-entry/include/form-line-item.html",
            currency_index="CURRENCY_INDEX",
            debit_credit="DEBIT_CREDIT",
            line_item_index="LINE_ITEM_INDEX",
            form=LineItemForm())


class CashReceiptJournalEntry(JournalEntryOperator):
    """A cash receipt journal entry."""
    CHECK_ORDER: int = 2
    """The order when checking the journal entry operator."""

    @property
    def form(self) -> Type[JournalEntryForm]:
        """Returns the form class.

        :return: The form class.
        """
        return CashReceiptJournalEntryForm

    def render_create_template(self, form: CashReceiptJournalEntryForm) -> str:
        """Renders the template for the form to create a journal entry.

        :param form: The journal entry form.
        :return: the form to create a journal entry.
        """
        return render_template(
            "accounting/journal-entry/receipt/create.html",
            form=form,
            journal_entry_type=JournalEntryType.CASH_RECEIPT,
            currency_template=self.__currency_template,
            line_item_template=self._line_item_template)

    def render_detail_template(self, journal_entry: JournalEntry) -> str:
        """Renders the template for the detail page.

        :param journal_entry: The journal entry.
        :return: the detail page.
        """
        return render_template("accounting/journal-entry/receipt/detail.html",
                               obj=journal_entry)

    def render_edit_template(self, journal_entry: JournalEntry,
                             form: CashReceiptJournalEntryForm) -> str:
        """Renders the template for the form to edit a journal entry.

        :param journal_entry: The journal entry.
        :param form: The form.
        :return: the form to edit a journal entry.
        """
        return render_template("accounting/journal-entry/receipt/edit.html",
                               journal_entry=journal_entry, form=form,
                               currency_template=self.__currency_template,
                               line_item_template=self._line_item_template)

    def is_my_type(self, journal_entry: JournalEntry) -> bool:
        """Checks and returns whether the journal entry belongs to the type.

        :param journal_entry: The journal entry.
        :return: True if the journal entry belongs to the type, or False
            otherwise.
        """
        return journal_entry.is_cash_receipt

    @property
    def __currency_template(self) -> str:
        """Renders and returns the template for the currency sub-form.

        :return: The template for the currency sub-form.
        """
        return render_template(
            "accounting/journal-entry/receipt/include/form-currency.html",
            currency_index="CURRENCY_INDEX",
            currency_code_data=default_currency_code(),
            credit_total="-")


class CashDisbursementJournalEntry(JournalEntryOperator):
    """A cash disbursement journal entry."""
    CHECK_ORDER: int = 1
    """The order when checking the journal entry operator."""

    @property
    def form(self) -> Type[JournalEntryForm]:
        """Returns the form class.

        :return: The form class.
        """
        return CashDisbursementJournalEntryForm

    def render_create_template(self, form: CashDisbursementJournalEntryForm) \
            -> str:
        """Renders the template for the form to create a journal entry.

        :param form: The journal entry form.
        :return: the form to create a journal entry.
        """
        return render_template(
            "accounting/journal-entry/disbursement/create.html",
            form=form,
            journal_entry_type=JournalEntryType.CASH_DISBURSEMENT,
            currency_template=self.__currency_template,
            line_item_template=self._line_item_template)

    def render_detail_template(self, journal_entry: JournalEntry) -> str:
        """Renders the template for the detail page.

        :param journal_entry: The journal entry.
        :return: the detail page.
        """
        return render_template(
            "accounting/journal-entry/disbursement/detail.html",
            obj=journal_entry)

    def render_edit_template(self, journal_entry: JournalEntry,
                             form: CashDisbursementJournalEntryForm) -> str:
        """Renders the template for the form to edit a journal entry.

        :param journal_entry: The journal entry.
        :param form: The form.
        :return: the form to edit a journal entry.
        """
        return render_template(
            "accounting/journal-entry/disbursement/edit.html",
            journal_entry=journal_entry, form=form,
            currency_template=self.__currency_template,
            line_item_template=self._line_item_template)

    def is_my_type(self, journal_entry: JournalEntry) -> bool:
        """Checks and returns whether the journal entry belongs to the type.

        :param journal_entry: The journal entry.
        :return: True if the journal entry belongs to the type, or False
            otherwise.
        """
        return journal_entry.is_cash_disbursement

    @property
    def __currency_template(self) -> str:
        """Renders and returns the template for the currency sub-form.

        :return: The template for the currency sub-form.
        """
        return render_template(
            "accounting/journal-entry/disbursement/include/form-currency.html",
            currency_index="CURRENCY_INDEX",
            currency_code_data=default_currency_code(),
            debit_total="-")


class TransferJournalEntry(JournalEntryOperator):
    """A transfer journal entry."""
    CHECK_ORDER: int = 3
    """The order when checking the journal entry operator."""

    @property
    def form(self) -> Type[JournalEntryForm]:
        """Returns the form class.

        :return: The form class.
        """
        return TransferJournalEntryForm

    def render_create_template(self, form: TransferJournalEntryForm) -> str:
        """Renders the template for the form to create a journal entry.

        :param form: The journal entry form.
        :return: the form to create a journal entry.
        """
        return render_template(
            "accounting/journal-entry/transfer/create.html",
            form=form,
            journal_entry_type=JournalEntryType.TRANSFER,
            currency_template=self.__currency_template,
            line_item_template=self._line_item_template)

    def render_detail_template(self, journal_entry: JournalEntry) -> str:
        """Renders the template for the detail page.

        :param journal_entry: The journal entry.
        :return: the detail page.
        """
        return render_template("accounting/journal-entry/transfer/detail.html",
                               obj=journal_entry)

    def render_edit_template(self, journal_entry: JournalEntry,
                             form: TransferJournalEntryForm) -> str:
        """Renders the template for the form to edit a journal entry.

        :param journal_entry: The journal entry.
        :param form: The form.
        :return: the form to edit a journal entry.
        """
        return render_template("accounting/journal-entry/transfer/edit.html",
                               journal_entry=journal_entry, form=form,
                               currency_template=self.__currency_template,
                               line_item_template=self._line_item_template)

    def is_my_type(self, journal_entry: JournalEntry) -> bool:
        """Checks and returns whether the journal entry belongs to the type.

        :param journal_entry: The journal entry.
        :return: True if the journal entry belongs to the type, or False
            otherwise.
        """
        return True

    @property
    def __currency_template(self) -> str:
        """Renders and returns the template for the currency sub-form.

        :return: The template for the currency sub-form.
        """
        return render_template(
            "accounting/journal-entry/transfer/include/form-currency.html",
            currency_index="CURRENCY_INDEX",
            currency_code_data=default_currency_code(),
            debit_total="-", credit_total="-")


JOURNAL_ENTRY_TYPE_TO_OP: dict[JournalEntryType, JournalEntryOperator] \
    = {JournalEntryType.CASH_RECEIPT: CashReceiptJournalEntry(),
       JournalEntryType.CASH_DISBURSEMENT: CashDisbursementJournalEntry(),
       JournalEntryType.TRANSFER: TransferJournalEntry()}
"""The map from the journal entry types to their operators."""


def get_journal_entry_op(journal_entry: JournalEntry,
                         is_check_as: bool = False) -> JournalEntryOperator:
    """Returns the journal entry operator that may be specified in the "as"
    query parameter.  If it is not specified, check the journal entry type from
    the journal entry.

    :param journal_entry: The journal entry.
    :param is_check_as: True to check the "as" parameter, or False otherwise.
    :return: None.
    """
    if is_check_as and "as" in request.args:
        type_dict: dict[str, JournalEntryType] \
            = {x.value: x for x in JournalEntryType}
        if request.args["as"] not in type_dict:
            abort(404)
        return JOURNAL_ENTRY_TYPE_TO_OP[type_dict[request.args["as"]]]
    for journal_entry_type in sorted(JOURNAL_ENTRY_TYPE_TO_OP.values(),
                                     key=lambda x: x.CHECK_ORDER):
        if journal_entry_type.is_my_type(journal_entry):
            return journal_entry_type



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The journal.

"""
import datetime as dt
from decimal import Decimal

import sqlalchemy as sa
from flask import render_template, Response
from sqlalchemy.orm import selectinload

from accounting.locale import gettext
from accounting.models import Currency, Account, JournalEntry, \
    JournalEntryLineItem
from accounting.report.period import Period, PeriodChooser
from accounting.report.utils.base_page_params import BasePageParams
from accounting.report.utils.base_report import BaseReport
from accounting.report.utils.csv_export import BaseCSVRow, csv_download, \
    period_spec
from accounting.report.utils.report_chooser import ReportChooser
from accounting.report.utils.report_type import ReportType
from accounting.report.utils.urls import journal_url
from accounting.utils.pagination import Pagination


class ReportLineItem:
    """A line item in the report."""

    def __init__(self, line_item: JournalEntryLineItem):
        """Constructs the line item in the report.

        :param line_item: The journal entry line item.
        """
        self.line_item: JournalEntryLineItem = line_item
        """The journal entry line item."""
        self.journal_entry: JournalEntry = line_item.journal_entry
        """The journal entry."""
        self.currency: Currency = line_item.currency
        """The account."""
        self.account: Account = line_item.account
        """The account."""
        self.description: str | None = line_item.description
        """The description."""
        self.debit: Decimal | None = line_item.debit
        """The debit amount."""
        self.credit: Decimal | None = line_item.credit
        """The credit amount."""
        self.amount: Decimal = line_item.amount
        """The amount."""


class CSVRow(BaseCSVRow):
    """A row in the CSV."""

    def __init__(self, journal_entry_date: str | dt.date,
                 currency: str,
                 account: str,
                 description: str | None,
                 debit: str | Decimal | None,
                 credit: str | Decimal | None,
                 note: str | None):
        """Constructs a row in the CSV.

        :param journal_entry_date: The journal entry date.
        :param currency: The currency.
        :param account: The account.
        :param description: The description.
        :param debit: The debit amount.
        :param credit: The credit amount.
        :param note: The note.
        """
        self.date: str | dt.date = journal_entry_date
        """The date."""
        self.currency: str = currency
        """The currency."""
        self.account: str = account
        """The account."""
        self.description: str | None = description
        """The description."""
        self.debit: str | Decimal | None = debit
        """The debit amount."""
        self.credit: str | Decimal | None = credit
        """The credit amount."""
        self.note: str | None = note
        """The note."""

    @property
    def values(self) -> list[str | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """
        return [self.date, self.currency, self.account, self.description,
                self.debit, self.credit, self.note]


class PageParams(BasePageParams):
    """The HTML page parameters."""

    def __init__(self, period: Period,
                 pagination: Pagination[JournalEntryLineItem],
                 line_items: list[JournalEntryLineItem]):
        """Constructs the HTML page parameters.

        :param period: The period.
        :param pagination: The pagination.
        :param line_items: The line items.
        """
        self.period: Period = period
        """The period."""
        self.pagination: Pagination[JournalEntryLineItem] = pagination
        """The pagination."""
        self.line_items: list[JournalEntryLineItem] = line_items
        """The line items."""
        self.period_chooser: PeriodChooser = PeriodChooser(
            lambda x: journal_url(x))
        """The period chooser."""

    @property
    def has_data(self) -> bool:
        """Returns whether there is any data on the page.

        :return: True if there is any data, or False otherwise.
        """
        return len(self.line_items) > 0

    @property
    def report_chooser(self) -> ReportChooser:
        """Returns the report chooser.

        :return: The report chooser.
        """
        return ReportChooser(ReportType.JOURNAL,
                             period=self.period)


def get_csv_rows(line_items: list[JournalEntryLineItem]) -> list[CSVRow]:
    """Composes and returns the CSV rows from the line items.

    :param line_items: The line items.
    :return: The CSV rows.
    """
    rows: list[CSVRow] = [CSVRow(gettext("Date"), gettext("Currency"),
                                 gettext("Account"), gettext("Description"),
                                 gettext("Debit"), gettext("Credit"),
                                 gettext("Note"))]
    rows.extend([CSVRow(x.journal_entry.date, x.currency.code,
                        str(x.account), x.description,
                        x.debit, x.credit, x.journal_entry.note)
                 for x in line_items])
    return rows


class Journal(BaseReport):
    """The journal."""

    def __init__(self, period: Period):
        """Constructs a journal.

        :param period: The period.
        """
        self.__period: Period = period
        """The period."""
        self.__line_items: list[JournalEntryLineItem] \
            = self.__query_line_items()
        """The line items."""

    def __query_line_items(self) -> list[JournalEntryLineItem]:
        """Queries and returns the line items.

        :return: The line items.
        """
        conditions: list[sa.BinaryExpression] = []
        if self.__period.start is not None:
            conditions.append(JournalEntry.date >= self.__period.start)
        if self.__period.end is not None:
            conditions.append(JournalEntry.date <= self.__period.end)
        return JournalEntryLineItem.query.join(JournalEntry)\
            .filter(*conditions)\
            .order_by(JournalEntry.date,
                      JournalEntry.no,
                      JournalEntryLineItem.is_debit.desc(),
                      JournalEntryLineItem.no)\
            .options(selectinload(JournalEntryLineItem.account),
                     selectinload(JournalEntryLineItem.currency),
                     selectinload(JournalEntryLineItem.journal_entry)).all()

    def csv(self) -> Response:
        """Returns the report as CSV for download.

        :return: The response of the report for download.
        """
        filename: str = f"journal-{period_spec(self.__period)}.csv"
        return csv_download(filename, get_csv_rows(self.__line_items))

    def html(self) -> str:
        """Composes and returns the report as HTML.

        :return: The report as HTML.
        """
        pagination: Pagination[JournalEntryLineItem] \
            = Pagination[JournalEntryLineItem](self.__line_items,
                                               is_reversed=True)
        params: PageParams = PageParams(period=self.__period,
                                        pagination=pagination,
                                        line_items=pagination.list)
        return render_template("accounting/report/journal.html",
                               report=params)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The period description composer.

"""
import datetime as dt

from accounting.locale import gettext


def get_desc(start: dt.date | None, end: dt.date | None) -> str:
    """Returns the period description.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The period description.
    """
    if start is None and end is None:
        return gettext("for all time")
    if start is None:
        return __get_until_desc(end)
    if end is None:
        return __get_since_desc(start)
    try:
        return __get_year_desc(start, end)
    except ValueError:
        pass
    try:
        return __get_month_desc(start, end)
    except ValueError:
        pass
    return __get_day_desc(start, end)


def __get_since_desc(start: dt.date) -> str:
    """Returns the description without the end day.

    :param start: The start of the period.
    :return: The description without the end day.
    """

    def get_start_desc() -> str:
        """Returns the description of the start day.

        :return: The description of the start day.
        """
        if start.month == 1 and start.day == 1:
            return str(start.year)
        if start.day == 1:
            return __format_month(start)
        return __format_day(start)

    return gettext("since %(start)s", start=get_start_desc())


def __get_until_desc(end: dt.date) -> str:
    """Returns the description without the start day.

    :param end: The end of the period.
    :return: The description without the start day.
    """

    def get_end_desc() -> str:
        """Returns the description of the end day.

        :return: The description of the end day.
        """
        if end.month == 12 and end.day == 31:
            return str(end.year)
        if (end + dt.timedelta(days=1)).day == 1:
            return __format_month(end)
        return __format_day(end)

    return gettext("until %(end)s", end=get_end_desc())


def __get_year_desc(start: dt.date, end: dt.date) -> str:
    """Returns the description as a year range.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The description as a year range.
    :raise ValueError: The period is not a year range.
    """
    if start.month != 1 or start.day != 1 \
            or end.month != 12 or end.day != 31:
        raise ValueError
    start_text: str = str(start.year)
    if start.year == end.year:
        return __get_in_desc(start_text)
    return __get_from_to_desc(start_text, str(end.year))


def __get_month_desc(start: dt.date, end: dt.date) -> str:
    """Returns the description as a month range.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The description as a month range.
    :raise ValueError: The period is not a month range.
    """
    if start.day != 1 or (end + dt.timedelta(days=1)).day != 1:
        raise ValueError
    start_text: str = __format_month(start)
    if start.year == end.year and start.month == end.month:
        return __get_in_desc(start_text)
    if start.year == end.year:
        return __get_from_to_desc(start_text, str(end.month))
    return __get_from_to_desc(start_text, __format_month(end))


def __get_day_desc(start: dt.date, end: dt.date) -> str:
    """Returns the description as a day range.

    :param start: The start of the period.
    :param end: The end of the period.
    :return: The description as a day range.
    :raise ValueError: The period is a month or year range.
    """
    start_text: str = __format_day(start)
    if start == end:
        return __get_in_desc(start_text)
    if start.year == end.year and start.month == end.month:
        return __get_from_to_desc(start_text, str(end.day))
    if start.year == end.year:
        end_month_day: str = f"{end.month}/{end.day}"
        return __get_from_to_desc(start_text, end_month_day)
    return __get_from_to_desc(start_text, __format_day(end))


def __format_month(month: dt.date) -> str:
    """Formats a month.

    :param month: The month.
    :return: The formatted month.
    """
    return f"{month.year}/{month.month}"


def __format_day(day: dt.date) -> str:
    """Formats a day.

    :param day: The day.
    :return: The formatted day.
    """
    return f"{day.year}/{day.month}/{day.day}"


def __get_in_desc(period: str) -> str:
    """Returns the description of a whole year, month, or day.

    :param period: The time period.
    :return: The description of a whole year, month, or day.
    """
    return gettext("in %(period)s", period=period)


def __get_from_to_desc(start: str, end: str) -> str:
    """Returns the description of a separated start and end.

    :param start: The start.
    :param end: The end.
    :return: The description of the separated start and end.
    """
    return gettext("in %(start)s-%(end)s", start=start, end=end)
